import pytest

from dictum_core.project import Project


def test_single_measure(project: Project):
    result = project.select(project.metrics.revenue).df()
    assert result.columns[0] == "revenue"
    assert result.iloc[0, 0] == 2328.6


def test_metric(project: Project):
    result = project.select(project.m.revenue_per_track).df()
    result.round(2).loc[0, "revenue_per_track"] == 0.66


def test_metric_groupby(project: Project):
    select = project.select(project.m.revenue_per_track).by(project.d.genre)
    result = select.df()
    assert result.shape == (25, 2)


def test_multiple_anchors(project: Project):
    result = project.select(project.metrics.revenue, project.metrics.track_count).df()
    assert next(result.itertuples()) == (0, 2328.6, 3503)


def test_multiple_anchors_by(project: Project, engine):
    select = project.select(project.m.revenue, project.m.track_count).by(
        project.d.genre
    )
    result = select.df()
    assert result.shape == (25, 3)


def test_groupby(project: Project):
    result = (
        project.select(project.metrics.revenue, project.metrics.track_count)
        .by(project.d.genre)
        .df()
    )
    assert result.shape == (25, 3)


def test_filter_eq(project: Project):
    result = (
        project.select(project.m.revenue, project.m.track_count)
        .by(project.d.genre)
        .where(project.d.artist == "Iron Maiden")
    ).df()
    assert result.shape == (4, 3)


def test_filter_ne(project: Project):
    result = (
        project.select(project.m.revenue, project.m.track_count)
        .by(project.d.artist)
        .where(project.d.genre != "Rock")
    ).df()
    assert result.shape == (165, 3)


def test_filter_gt(project: Project):
    result = project.select(project.m.revenue).where(project.d.order_amount > 5).df()
    assert result.iloc[0, 0] == 1797.81


def test_filter_ge(project: Project):
    result = project.select(project.m.revenue).where(project.d.order_amount >= 5).df()
    assert result.iloc[0, 0] == 1797.81


def test_filter_lt(project: Project):
    result = project.select(project.m.revenue).where(project.d.order_amount < 5).df()
    assert result.iloc[0, 0] == 530.79


def test_filter_le(project: Project):
    result = project.select(project.m.revenue).where(project.d.order_amount <= 5).df()
    assert result.iloc[0, 0] == 530.79


def test_filter_isin(project: Project):
    result = (
        project.select(project.m.revenue)
        .where(project.d.genre.isin("Alternative", "Rock"))
        .by(project.d.genre)
        .df()
    )
    assert result.shape == (2, 2)


def test_date_unit(project: Project):
    result = project.select(project.m.revenue).by(project.d.invoice_date.year).df()
    assert result.shape == (5, 2)


def test_step_transform(project: Project):
    result = project.select(project.m.revenue).by(project.d.order_amount.step(10)).df()
    assert result.shape == (3, 2)


def test_dimension_alias(project: Project):
    result = (
        project.select(project.m.revenue)
        .by(project.d.invoice_date.year.name("year"))
        .df()
    )
    assert tuple(result.columns) == ("year", "revenue")


def test_metric_alias(project: Project):
    result = project.select(project.m.revenue.name("test")).by(project.d.genre).df()
    assert tuple(result.columns) == ("genre", "test")


def test_datetrunc_and_inrange(project: Project):
    result = (
        project.select(project.m.revenue)
        .by(project.d.invoice_date.datetrunc("week"))
        .where(project.d.invoice_date.inrange("2010-01-01", "2011-12-31"))
        .df()
    )
    assert result.shape == (81, 2)


def test_top_with_measure_basic(project: Project):
    result = (
        project.select(project.m.revenue)
        .by(project.d.genre)
        .limit(project.m.revenue.top(5))
        .df()
    )
    assert set(result.genre) == {
        "Metal",
        "Latin",
        "Rock",
        "Alternative & Punk",
        "TV Shows",
    }


def test_top_with_metrics_basic(project: Project):
    result = (
        project.select(project.m.revenue)
        .by(project.d.genre)
        .limit(project.m.revenue_per_track.top(5))
        .df()
    )
    assert set(result.genre) == {
        "Bossa Nova",
        "Sci Fi & Fantasy",
        "TV Shows",
        "Comedy",
        "Science Fiction",
    }


def test_top_with_multiple_basic(project: Project):
    result = (
        project.select(project.m.revenue)
        .by(project.d.genre)
        .limit(
            project.m.revenue.top(10),
            project.m.revenue_per_track.top(10),
        )
        .df()
    )
    assert set(result.genre) == {
        "Blues",
        "TV Shows",
        "Drama",
        "Alternative & Punk",
        "Metal",
        "R&B/Soul",
    }


def test_top_with_multiple_basic_reverse_order(project: Project):
    result = (
        project.select(project.m.revenue)
        .by(project.d.genre)
        .limit(
            project.m.revenue_per_track.top(10),
            project.m.revenue.top(10),
        )
        .df()
    )
    assert set(result.genre) == {
        "Blues",
        "TV Shows",
        "Drama",
        "Alternative & Punk",
        "Metal",
        "R&B/Soul",
    }


def test_top_with_measure_basic_metric(project: Project):
    result = (
        project.select(project.m.revenue_per_track)
        .by(project.d.genre)
        .limit(project.m.revenue.top(5))
        .df()
    )
    assert set(result.genre) == {
        "Metal",
        "Latin",
        "Rock",
        "Alternative & Punk",
        "TV Shows",
    }


def test_top_with_metric_basic_metric(project: Project):
    result = (
        project.select(project.m.revenue_per_track)
        .by(project.d.genre)
        .limit(project.m.revenue_per_track.top(5))
        .df()
    )
    assert set(result.genre) == {
        "Bossa Nova",
        "Sci Fi & Fantasy",
        "TV Shows",
        "Comedy",
        "Science Fiction",
    }


def test_top_with_multiple_basic_metric(project: Project):
    result = (
        project.select(project.m.revenue_per_track)
        .by(project.d.genre)
        .limit(
            project.m.revenue.top(10),
            project.m.revenue_per_track.top(10),
        )
        .df()
    )
    assert set(result.genre) == {
        "Blues",
        "TV Shows",
        "Drama",
        "Alternative & Punk",
        "Metal",
        "R&B/Soul",
    }


def test_top_with_multiple_basic_metric_reverse_order(project: Project):
    result = (
        project.select(project.m.revenue_per_track)
        .by(project.d.genre)
        .limit(
            project.m.revenue_per_track.top(10),
            project.m.revenue.top(10),
        )
        .df()
    )
    assert set(result.genre) == {
        "Blues",
        "TV Shows",
        "Drama",
        "Alternative & Punk",
        "Metal",
        "R&B/Soul",
    }


def test_top_with_measure_within(project: Project):
    result = (
        project.select(project.m.revenue)
        .by(project.d.genre, project.d.artist)
        .limit(project.m.revenue.top(3, within=[project.d.genre]))
        .df()
    )
    assert result.shape == (56, 3)


def test_top_with_measure_within_of(project: Project):
    result = (
        project.select(project.m.revenue)
        .by(project.d.genre, project.d.artist)
        .limit(
            project.m.revenue.top(3, within=[project.d.genre]),
            project.m.revenue.top(3, of=[project.d.genre]),
        )
        .df()
    )
    assert result.shape == (9, 3)
    assert set(result.genre) == {"Latin", "Metal", "Rock"}


def test_top_with_metric_within_of(project: Project):
    result = (
        project.select(project.m.revenue, project.m.revenue_per_track)
        .by(project.d.genre, project.d.artist)
        .limit(
            project.m.revenue_per_track.top(3, within=[project.d.genre]),
            project.m.revenue_per_track.top(3, of=[project.d.genre]),
        )
        .df()
    )
    assert result.shape == (6, 4)


def test_tops_with_total(project: Project, engine):
    result = (
        project.select("revenue", "revenue.total")
        .by("customer_country", "customer_city")
        .limit(
            "revenue.top(3) of (customer_country)",
            "revenue.top(1) of (customer_city) within (customer_country)",
        )
        .df()
    )
    assert result.shape == (3, 4)


def test_tops_with_total_within(project: Project):
    result = project.ql(
        """
    select revenue, revenue.total within (customer_country)
    by customer_country, customer_city
    limit revenue.top(5) of (customer_country),
        revenue.top(1) of (customer_city) within (customer_country)
    """
    ).df()
    assert result.shape == (5, 4)


def test_tops_with_matching_total_and_percent(project: Project):
    result = project.ql(
        """
    select revenue,
        revenue.percent of (customer_city) as "% of City Revenue",
        revenue.total within (customer_country) as "Total Revenue: Country"
    by customer_country, customer_city
    limit revenue.top(5) of (customer_country),
        revenue.top(1) of (customer_city) within (customer_country)
    """
    ).df()
    assert result.shape == (5, 5)


def test_total_basic(project: Project):
    result = (
        project.select(project.m.revenue, project.m.revenue.total())
        .by(project.d.genre)
        .df()
    )
    assert result.shape == (24, 3)
    assert result.revenue__total.unique().tolist() == [2328.6]


def test_total_metric(project: Project):
    select = (
        project.select(
            project.m.revenue_per_track,
            project.m.revenue_per_track.total(within=[project.d.genre]),
        )
        .by(project.d.genre, project.d.artist)
        .limit(project.m.revenue.top(3, within=[project.d.genre]))
    )
    result = select.df()
    assert result.shape == (56, 4)
    assert len(set(result.genre)) == len(
        set(result.revenue_per_track__total_within_genre)
    )


def test_total_transformed_dimension(project: Project):
    result = (
        project.select("revenue.total within (Time.year)")
        .by("Time.year", "Time.month")
        .df()
    )
    assert len(result.iloc[:, -1].unique()) == 5


def test_total_filters(project: Project):
    """Test that filters are applied to the table transforms too (was a bug)"""
    select = (
        project.select(project.m.revenue.total())
        .by(project.d.genre)
        .where(project.d.genre.isin("Rock", "Alternative & Punk"))
    )
    result = select.df()
    assert result["revenue__total"].unique()[0] == 1068.21


def test_percent_basic(project: Project):
    result = project.select(project.m.revenue.percent()).by(project.d.genre).df()
    assert result.shape == (24, 2)
    assert result["revenue__percent"].sum() == 1


def test_percent_of(project: Project):
    result = project.select("revenue.percent of (artist)").by("genre", "artist").df()
    unique = result.groupby("genre").sum().iloc[:, 0].round(4).unique()
    assert unique.size == 1
    assert unique[0] == 1.0


def test_percent_within(project: Project):
    result = project.select("revenue.percent within (genre)").by("genre", "artist").df()
    unique = result.groupby("genre").sum().iloc[:, 0].round(4).unique()
    assert unique.size == 1
    assert unique[0] == 1.0


def test_percent_of_within(project: Project):
    result = (
        project.select(
            "revenue",
            "revenue.percent of (artist) within (genre)",
        ).by("genre", "artist", "album")
        # .where("artist = 'Faith No More'")
        .df()
    )
    values = result.query("genre == 'TV Shows' and artist == 'The Office'")[
        "revenue__percent_of_artist_within_genre"
    ].unique()
    assert len(values) == 1
    assert values.round(4)[0] == 0.3404


def test_percent_with_top(project: Project):
    """Percent should be calculated before top"""
    result = (
        project.select(project.m.revenue.percent())
        .by(project.d.genre)
        .limit(project.m.revenue.top(5))
        .df()
    )
    assert result.shape == (5, 2)
    assert round(result["revenue__percent"].sum(), 4) == 0.7752


def test_percent_with_top_within(project: Project):
    result = (
        project.select(
            project.m.revenue.percent(within=[project.d.genre]).name(
                "artist_revenue_within_genre"
            )
        )
        .by(project.d.genre, project.d.artist)
        .limit(project.m.revenue.top(1, within=[project.d.genre]))
        .df()
    )
    assert result.shape == (24, 3)
    assert result.columns[-1] == "artist_revenue_within_genre"  # alias works


def test_percent_integer(project: Project):
    """Check that the output type is handled correctly (float, not the original int)
    and that the percentages for unique values do not add up to 100%"""
    select = project.select(project.m.unique_paying_customers.percent()).by(
        project.d.genre
    )
    result = select.df()
    assert round(result["unique_paying_customers__percent"].sum(), 2) == 7.46


def test_format_metric(project: Project):
    result = project.select(project.m.revenue).df(format=True)
    assert result.columns[0] == "Revenue"  # column name comes from the metric name
    assert result.iloc[0, 0] == "$2,328.60"  # the value is formatted


def test_format_dimension_name(project: Project):
    result = project.select(project.m.revenue).by(project.d.genre).df(format=True)
    assert list(result.columns) == ["Genre", "Revenue"]


def test_format_transform(project: Project):
    result = (
        project.select(project.m.revenue)
        .by(project.d.invoice_date.year)
        .df(format=True)
    )
    assert list(result.columns) == ["Invoice Date (year)", "Revenue"]
    assert result.iloc[0, 0] == "2009"


def test_format_dimension_transform_alias(project: Project):
    result = (
        project.select(project.m.revenue.percent().name("Percent of Revenue"))
        .by(project.d.invoice_date.year.name("Year"))
        .df(format=True)
    )
    assert list(result.columns) == ["Year", "Percent of Revenue"]


def test_filtered_table(project: Project):
    result = project.select(project.m.rock_revenue).df()
    assert result.iloc[0, 0] == 826.65


def test_filtered_measure(project: Project):
    result = project.select("music_revenue").df()
    assert result.iloc[0, 0] == 2107.71


def test_filtered_and_unfiltered_measures_together(project: Project):
    result = project.select("revenue", "music_revenue").df()
    assert next(result.itertuples(index=False)) == (2328.6, 2107.71)


def test_generic_time(project: Project):
    result = project.select("revenue").by("Time.datetrunc('year')").df()
    assert result.shape == (5, 2)


def test_generic_time_alias_display_name(project: Project):
    result = project.select("revenue").by("Year as test").df(format=True)
    assert result.columns[0] == "test"


def test_generic_time_format(project: Project):
    result = project.select("revenue").by("Year").df(format=True)
    assert result.iloc[0]["Year"] == "2009"


@pytest.mark.parametrize(
    "dimension,n",
    [
        ("Year", 5),
        ("Quarter", 20),
        ("Month", 60),
        ("Week", 202),
        ("Day", 354),
        ("Date", 354),
        ("Hour", 354),  # dates in the database
        ("Minute", 354),
        ("Second", 354),
        ("Time", 354),
    ],
)
def test_generic_time_trunc(project: Project, dimension: str, n: int):
    result = project.select("revenue").by(dimension).df()
    assert result.shape == (n, 2)


def test_generic_time_trunc_transform(project: Project):
    result = project.select("revenue").by("Month.datetrunc('year')").df()
    assert result.shape == (5, 2)


def test_sum_table_transform(project: Project):
    result = project.select("revenue.sum").by("genre").df()
    assert result["revenue__sum"].unique().size == 1
    assert result["revenue__sum"].unique()[0] == 2328.6


def test_sum_table_transform_within(project: Project):
    result = (
        project.select("revenue", "revenue.sum() within (genre)")
        .by("genre", "artist")
        .df()
    )
    gb = result.groupby("genre").aggregate(
        {"revenue": "sum", "revenue__sum_within_genre": "max"}
    )
    assert (gb["revenue"].round(2) == gb["revenue__sum_within_genre"]).all()


def test_measure_with_related_column(project: Project):
    """Test that related columns are supported in measures"""
    project.select("unique_paying_customers").df()  # no error


def test_default_time_format(project: Project):
    result = project.select("revenue").by("invoice_date").df(format=True)
    assert result.iloc[0]["Invoice Date"] == "1/1/09"


def test_percents_without_alias(project: Project):
    """Bug found writing the docs. Something is wrong when running this query, fails
    in the pandas merge Visitor with KeyError, column not found.
    """
    (
        project.select(
            "revenue.percent within (invoice_date.year)",
            "revenue.percent of (invoice_date.quarter) within (invoice_date.year)",
        )
        .by("invoice_date.year", "invoice_date.quarter", "invoice_date.month")
        .df()
    )


def test_total_of_within_keyerror(project: Project):
    """Bug found writing the docs, similar to the above"""
    (
        project.select(
            "revenue.total within (Year)",
            "revenue.total of (Year)",
        ).by("Year", "Quarter")
    ).df()



import altair as alt

from dictum_core.project.altair.data import DictumData
from dictum_core.project.altair.monkeypatch import (
    is_dictum_definition,
    monkeypatch_altair,
    request_from_field,
    requests_from_channel,
)
from dictum_core.project.project import Project
from dictum_core.schema.query import Query

monkeypatch_altair()


def test_encode_shorthand(project: Project):
    assert (
        project.chart().encode(x=project.m.revenue).encoding.x.field == "metric:revenue"
    )


def test_channel_init(project: Project):
    channel = alt.X(project.m.revenue)
    assert channel.shorthand == alt.Undefined
    assert channel.field == "metric:revenue"


def test_facet_resolve(project: Project):
    chart = project.chart().mark_bar().encode(x=project.m.revenue)
    media_type = alt.FieldName("dimension:media_type")
    assert chart.facet(project.d.media_type)._kwds["facet"].field == media_type
    assert chart.facet(row=project.d.media_type)._kwds["facet"].row.field == media_type
    assert (
        chart.facet(column=project.d.media_type)._kwds["facet"].column.field
        == media_type
    )


def test_repeat_resolve(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(x=project.d.media_type, y=alt.repeat("repeat"))
    )
    assert chart.repeat([project.m.revenue])._kwds["repeat"][0] == "metric:revenue"
    assert (
        chart.repeat(row=[project.m.revenue])._kwds["repeat"].row[0] == "metric:revenue"
    )
    assert (
        chart.repeat(column=[project.m.revenue])._kwds["repeat"].column[0]
        == "metric:revenue"
    )
    assert (
        chart.repeat(layer=[project.m.revenue])._kwds["repeat"].layer[0]
        == "metric:revenue"
    )


def test_iterunits():
    ch = alt.Chart("#").mark_bar()

    assert len(list(ch._iterunits())) == 1
    assert len(list((ch + ch)._iterunits())) == 2
    assert len(list((ch | ch)._iterunits())) == 2
    assert len(list((ch & ch)._iterunits())) == 2
    assert len(list((ch | ch | ch)._iterunits())) == 3
    assert len(list((ch & ch & ch)._iterunits())) == 3
    assert len(list((ch | ch & ch)._iterunits())) == 3
    assert len(list((ch & ch | ch)._iterunits())) == 3
    assert len(list((ch & ch | ch & ch)._iterunits())) == 4
    assert len(list((ch | ch & ch | ch)._iterunits())) == 4
    assert len(list(((ch | ch) & ch)._iterunits())) == 3
    assert len(list((ch | (ch & ch))._iterunits())) == 3

    facet = alt.Chart("#").mark_bar().facet(column="test:N")
    assert len(list(facet._iterunits())) == 1
    assert len(list((ch | (ch & facet))._iterunits())) == 3

    repeat = alt.Chart("#").mark_bar().repeat(column=["test"])
    assert len(list(repeat._iterunits())) == 1
    assert len(list((ch | (ch & repeat))._iterunits())) == 3


def test_iterunits_data():
    a = alt.NamedData("a")
    b = alt.NamedData("b")
    ch = alt.Chart(a).mark_bar()

    assert next(ch._iterunits())[1] == a
    assert [d for _, d in (ch | ch)._iterunits()] == [a, a]
    assert [d for _, d in (ch & ch)._iterunits()] == [a, a]
    assert [d for _, d in (ch & ch.properties(data=b))._iterunits()] == [a, b]
    assert [d for _, d in (ch & (ch.properties(data=b) & ch))._iterunits()] == [a, b, a]

    facet = ch.facet(column="test:N")
    assert [d for _, d in facet._iterunits()] == [a]
    assert [d for _, d in (ch | facet.properties(data=b))._iterunits()] == [a, b]

    repeat = alt.Chart(a).mark_bar().encode(x="x:N", y="y:Q").repeat(["a"])
    assert list(repeat._iterunits()) == [(repeat, a)]


def test_iterchannels():
    ch = alt.Chart("#").mark_bar().encode(x="x:Q", y="y:N")
    assert len(list(ch._iterchannels())) == 2
    assert len(list(ch.encode(text="z:N")._iterchannels())) == 3
    assert len(list(ch.facet(column="test:N").spec._iterchannels())) == 2
    assert len(list(ch.facet(column="test:N")._iterchannels())) == 2


def test_chart_to_dict(project: Project):
    d = (
        project.chart()
        .mark_bar(stroke="white")
        .encode(
            x=project.m.track_count,
            y=alt.Y(project.d.genre, sort="-x"),
            color=alt.Color(project.d.media_type, sort="-x"),
            order=alt.Order(project.m.track_count, sort="descending"),
        )
    ).to_dict()
    assert d["encoding"]["x"]["field"] == "metric:track_count"
    assert d["encoding"]["y"]["field"] == "dimension:genre"


def test_chart_query(project: Project):
    chart = (
        project.chart()
        .mark_bar(stroke="white")
        .encode(
            x=project.d.invoice_date.year,
            y=project.m.revenue,
            color=alt.Color(project.d.media_type),
        )
    )
    assert chart._query() == Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "revenue"}}],
            "dimensions": [
                {"dimension": {"id": "media_type"}},
                {"dimension": {"id": "invoice_date", "transforms": [{"id": "year"}]}},
            ],
        }
    )


def test_query_sort(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(x=alt.X(project.d.genre, sort=project.m.revenue))
    )
    assert chart._query() == Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "revenue"}}],
            "dimensions": [{"dimension": {"id": "genre"}}],
        }
    )


def test_repeat_not_resolving_data(project: Project):
    chart = (
        project.chart()
        .mark_line()
        .encode(x=project.d.genre, y=alt.Y(alt.repeat("layer")))
        .repeat(layer=[project.m.track_count, project.m.revenue])
    )
    assert len(list(chart._iterunits())) > 0
    unit, data = next(chart._iterunits())
    assert unit is chart
    assert isinstance(data, DictumData)


def test_channel_sort(project: Project):
    assert alt.X(project.d.genre, sort=project.m.revenue).sort.field == "metric:revenue"


def test_is_dictum_definition():
    assert is_dictum_definition(alt.FieldName("metric:revenue"))
    assert is_dictum_definition("metric:revenue")
    assert is_dictum_definition("dimension:x")
    assert is_dictum_definition("dimension:x.test.blah()")
    assert not is_dictum_definition("blahblah:Q")


def test_request_from_field():
    assert request_from_field("metric:revenue").metric.id == "revenue"
    assert request_from_field(alt.FieldName("metric:revenue")).metric.id == "revenue"
    assert request_from_field(alt.FieldName("blah")) is None
    assert request_from_field("dimension:xx.date").dimension.id == "xx"
    assert request_from_field("dimension:xx.date").dimension.transforms[0].id == "date"


def test_requests_from_channel(project: Project):
    assert len(requests_from_channel(alt.X(field="metric:test"))) == 1
    assert (
        len(
            requests_from_channel(
                alt.X(
                    project.d.genre,
                    sort=alt.EncodingSortField(project.m.revenue, order="descending"),
                )
            )
        )
        == 2
    )
    assert len(requests_from_channel(alt.X("something:Q"))) == 0



from typing import Dict

import dictum_core.model
import dictum_core.project
from dictum_core.engine.metrics import limit_transforms
from dictum_core.engine.metrics import transforms as table_transforms
from dictum_core.model.scalar import transforms as scalar_transforms
from dictum_core.project.altair.encoding import AltairEncodingChannelHook
from dictum_core.project.templates import environment
from dictum_core.schema.query import (
    QueryDimension,
    QueryDimensionRequest,
    QueryMetric,
    QueryMetricRequest,
    QueryScalarTransform,
    QueryTableTransform,
)

scalar_transforms = set(scalar_transforms)
table_transforms = set(table_transforms) | set(limit_transforms)


class ProjectCalculation(AltairEncodingChannelHook):
    kind: str

    def __init__(self, calculation, locale: str):
        self.calculation = calculation
        self.locale = locale

    def encoding_fields(self, cls=None) -> dict:
        return {"field": f"{self._type}:{self}"}

    @property
    def _type(self) -> str:
        if isinstance(self.calculation, dictum_core.model.Metric):
            return "metric"
        return "dimension"

    def _repr_html_(self):
        template = environment.get_template("calculation.html.j2")
        calculation = self.calculation
        if (
            isinstance(calculation, dictum_core.model.Metric)
            and len(calculation.measures) == 1
        ):
            calculation = calculation.measures[0]
        prefix = "$" if self._type == "metric" else ":"
        return template.render(calculation=calculation, prefix=prefix)

    def name(self, name: str):
        self.request.alias = name
        return self

    def __eq__(self, other):
        return self.eq(other)

    def __ne__(self, other):
        return self.ne(other)

    def __gt__(self, other):
        return self.gt(other)

    def __ge__(self, other):
        return self.ge(other)

    def __lt__(self, other):
        return self.lt(other)

    def __le__(self, other):
        return self.le(other)

    def __invert__(self):
        return self.invert()

    def __str__(self):
        return self.request.render()


class ProjectMetricRequest(ProjectCalculation):
    kind = "metric"

    def __init__(self, calculation, locale: str):
        self.request = QueryMetricRequest(metric=QueryMetric(id=calculation.id))
        super().__init__(calculation, locale)

    def __getattr__(self, name: str):
        if name not in table_transforms:
            raise AttributeError(name)
        self.request.metric.transforms.append(QueryTableTransform(id=name))
        return self

    def __dir__(self):
        return table_transforms.keys()

    def __call__(self, *args, of=None, within=None):
        of = [] if of is None else of
        within = [] if within is None else within
        self.request.metric.transforms[-1].args = list(args)
        self.request.metric.transforms[-1].of = [i.request for i in of]
        self.request.metric.transforms[-1].within = [i.request for i in within]
        return self


class ProjectMetric(ProjectMetricRequest):
    def name(self, name: str):
        return ProjectMetricRequest(
            calculation=self.calculation, locale=self.locale
        ).name(name)

    def __getattr__(self, name: str):
        if name not in table_transforms:
            raise AttributeError(name)
        return getattr(
            ProjectMetricRequest(calculation=self.calculation, locale=self.locale),
            name,
        )


class ProjectMetrics:
    def __init__(self, project: "dictum_core.project.Project"):
        self.__project = project
        self.__metrics: Dict[str, ProjectMetric] = {
            m.id: ProjectMetric(m, project.model.locale)
            for m in project.model.metrics.values()
        }

    def __getattr__(self, attr: str) -> ProjectMetric:
        return self.__metrics[attr]

    def __getitem__(self, key: str) -> ProjectMetric:
        return self.__metrics[key]

    def __dir__(self):
        return self.__metrics.keys()

    def _repr_html_(self):
        template = environment.get_template("calculations.html.j2")
        return template.render(calculations=self.__project.model.metrics.values())


class ProjectDimensionRequest(ProjectCalculation):
    kind = "dimension"

    def __init__(self, calculation, locale: str):
        self.request = QueryDimensionRequest(
            dimension=QueryDimension(id=calculation.id)
        )
        super().__init__(calculation, locale)

    def __getattr__(self, name: str):
        if name not in scalar_transforms:
            raise AttributeError(name)  # for Jupyter checking for _repr_html_ etc.
        self.request.dimension.transforms.append(QueryScalarTransform(id=name))
        return self

    def __call__(self, *args):
        self.request.dimension.transforms[-1].args = list(args)
        return self


class ProjectDimension(ProjectDimensionRequest):
    def __getattr__(self, name: str):
        if name not in scalar_transforms:
            raise AttributeError(name)  # for Jupyter checking for _repr_html_ etc.
        return getattr(
            ProjectDimensionRequest(self.calculation, locale=self.locale),
            name,
        )

    def name(self, name: str):
        return ProjectDimensionRequest(self.calculation, locale=self.locale).name(name)

    def __str__(self):
        return self.calculation.id


class ProjectDimensions:
    def __init__(self, project: "dictum_core.project.Project"):
        self.__project = project
        self.__dimensions: Dict[str, ProjectDimension] = {
            d.id: ProjectDimension(d, project.model.locale)
            for d in project.model.dimensions.values()
        }

    def __getattr__(self, attr: str) -> ProjectDimension:
        return self.__dimensions[attr]

    def __getitem__(self, key: str) -> ProjectDimension:
        return self.__dimensions[key]

    def _repr_html_(self):
        template = environment.get_template("calculations.html.j2")
        return template.render(calculations=self.__project.model.dimensions.values())



from dictum_core.schema.model import Model
from dictum_core.schema.model.calculations import (
    Dimension,
    DimensionsUnion,
    Measure,
    Metric,
)
from dictum_core.schema.model.format import FormatConfig
from dictum_core.schema.model.table import Table
from dictum_core.schema.model.type import Type
from dictum_core.schema.project import Project
from dictum_core.schema.query import (
    Query,
    QueryDimension,
    QueryDimensionRequest,
    QueryMetric,
    QueryMetricRequest,
    QueryScalarTransform,
    QueryTableTransform,
    QueryTransform,
)

__all__ = [
    "Dimension",
    "DimensionsUnion",
    "FormatConfig",
    "Measure",
    "Metric",
    "Model",
    "Project",
    "Query",
    "QueryDimension",
    "QueryDimensionRequest",
    "QueryMetric",
    "QueryMetricRequest",
    "QueryScalarTransform",
    "QueryTableTransform",
    "QueryTransform",
    "Table",
    "Type",
]



import logging
from functools import cached_property
from typing import Any, Dict, List, Optional, Union

import sqlparse
from lark import Transformer, Tree
from pandas import DataFrame, read_sql
from sqlalchemy import (
    Date,
    DateTime,
    Float,
    Integer,
    MetaData,
    Table,
    and_,
    case,
    cast,
    create_engine,
    distinct,
    func,
    not_,
    or_,
    select,
    true,
)
from sqlalchemy.engine.url import URL
from sqlalchemy.sql import Select
from sqlalchemy.sql.functions import coalesce

import dictum_core.model
from dictum_core.backends.base import Backend, Compiler
from dictum_core.backends.mixins.arithmetic import ArithmeticCompilerMixin
from dictum_core.engine import Column, LiteralOrderItem, RelationalQuery

logger = logging.getLogger(__name__)


def get_case_insensitive_column(obj, column: str):
    """SQL is case-insensitive, but SQLAlchemy isn't, because columns are stored
    in a dict-like structure, keys exactly as specified in the database. That's
    why this function is needed.
    """
    columns = obj.selected_columns if isinstance(obj, Select) else obj.columns
    if column in columns:
        return columns[column]
    for k, v in columns.items():
        if k.lower() == column.lower():
            return v
    raise KeyError(f"Can't find column {column} in {obj.name}")


class ColumnTransformer(Transformer):
    """Replaces columns in the expressions with actual column objects from the join."""

    def __init__(self, tables, visit_tokens: bool = True) -> None:
        self._tables = tables
        super().__init__(visit_tokens=visit_tokens)

    def column(self, children: list):
        identity, field = children
        return get_case_insensitive_column(self._tables[identity], field)


class SQLAlchemyCompiler(ArithmeticCompilerMixin, Compiler):
    def __init__(self, backend: "SQLAlchemyBackend"):
        self.backend = backend
        super().__init__()

    def column(self, _):
        """A no-op. SQLAlchemy columns need to know about tables,
        so this transformation is done beforehand with ColumnTransformer.
        """

    def IN(self, a, b):
        return a.in_(b)

    def NOT(self, x):
        return not_(x)

    def AND(self, a, b):
        return and_(a, b)

    def OR(self, a, b):
        return or_(a, b)

    def isnull(self, value):
        return value == None  # noqa: E711

    def case(self, whens, else_=None):
        return case(whens, else_=else_)

    def sum(self, arg):
        return func.sum(arg)

    def avg(self, arg):
        return func.avg(arg)

    def min(self, arg):
        return func.min(arg)

    def max(self, arg):
        return func.max(arg)

    def count(self, arg=None):
        args = []
        if arg is not None:
            args = [arg]
        return func.count(*args)

    def countd(self, arg):
        return func.count(distinct(arg))

    def call_window(
        self, fn: str, args: list, partition: list, order: list, rows: list
    ):
        if order:
            order_by = []
            for item in order:
                col, asc = item.children
                col = col.asc() if asc else col.desc()
                order_by.append(col)
            order = order_by
        return super().call_window(fn, *args, partition, order, rows)

    def window_sum(self, arg, partition, order, rows):
        return func.sum(arg).over(partition_by=partition, order_by=order)

    def window_row_number(self, _, partition, order, rows):
        return func.row_number().over(partition_by=partition, order_by=order)

    def floor(self, arg):
        return func.floor(arg)

    def ceil(self, arg):
        return func.ceil(arg)

    def abs(self, arg):
        return func.abs(arg)

    def datepart(self, part, arg):
        return func.date_part(part, arg)

    def datetrunc(self, part, arg):
        return func.date_trunc(part, arg)

    def datediff(self, part, start, end):
        return func.datediff(part, start, end)

    def now(self):
        return func.now()

    def today(self):
        return self.todate(self.now())

    def coalesce(self, *args):
        return func.coalesce(*args)

    def tointeger(self, arg):
        return cast(arg, Integer)

    def tofloat(self, arg):
        return cast(arg, Float)

    def todate(self, arg):
        return cast(arg, Date)

    def todatetime(self, arg):
        return cast(arg, DateTime)

    def _table(self, source: Union[str, Dict]):
        if isinstance(source, str):
            return self.backend.table(source)
        if isinstance(source, dict):
            schema = source.get("schema")
            table = source.get("table")
            if table is None:
                raise ValueError(f"table is required for a {self.type} backend")
            return self.backend.table(table, schema)
        raise ValueError(f"Source must be a str or a dict for a {self.type} backend")

    def compile_query(self, query: RelationalQuery):
        if isinstance(query.source, dictum_core.model.Table):
            id_ = query.source.id
            table = self._table(query.source.source)
        elif isinstance(query.source, RelationalQuery):
            id_ = None
            table = self.compile_query(query.source)

        tables = {id_: table}

        # replaces column refs with actual SQLA column objs
        column_transformer = ColumnTransformer(tables)

        # join the tables
        for join in query.joins:
            if isinstance(join.right, RelationalQuery):
                right_table = self.compile_query(join.right)
            else:
                right_table = self._table(join.right.source)
            right_table = right_table.alias(join.right_identity)
            tables[join.right_identity] = right_table
            join_expr = self.transformer.transform(
                column_transformer.transform(join.expr)
            )
            table = table.join(right_table, join_expr, isouter=(not join.inner))

        # add calcs
        columns = {}
        for column in query.columns:
            columns[column.name] = self.transformer.transform(
                column_transformer.transform(column.expr)
            )
        groupby = []
        for expr in query.groupby:
            compiled = self.transformer.transform(column_transformer.transform(expr))
            groupby.append(compiled)

        stmt = (
            select(*(v.label(k) for k, v in columns.items()))
            .select_from(table)
            .group_by(*groupby)
        )

        # apply filters
        for expr in query.filters:
            condition = self.transformer.transform(column_transformer.transform(expr))
            stmt = stmt.where(condition)

        # apply order and limit
        # order only makes sense at this stage if there's a limit
        if query.limit and query.order:
            order_clauses = []
            for item in query.order:
                expr = self.transformer.transform(
                    column_transformer.transform(item.expr)
                )
                if not item.ascending:
                    expr = expr.desc()
                order_clauses.append(expr)
            stmt = stmt.order_by(*order_clauses).limit(query.limit)

        return stmt

    def merge_queries(self, queries: List[Select], merge_on: List[str]):
        """Merge multiple queries. Wrap each in subquery and then full outer join them.
        The join condition is constructed by chaining the coalesced statements:

        If A and B are dimension columns and x, y, z are measures:

        SELECT coalesce(coalesce(t1.A, t2.A), t3.A) as A,
            coalesce(coalesce(t1.b, t2.B), t3.B) as B,
            t1.x, t2.y, t3.z
        FROM (...) as t1
        FULL OUTER JOIN (...) as t2
            ON t1.A = t2.A
            AND t1.B = t2.B
        FULL OUTER JOIN (...) as t3
            ON coalesce(t1.A, t2.A) = t3.A
            AND coalesce(t1.B, t2.B) = t3.B
        """
        subqueries = [q.subquery() for q in queries]
        joined, *joins = subqueries

        coalesced = {k: v for k, v in joined.c.items() if k in merge_on}

        for join in joins:
            # join on columns that are common between merge_on, the table that's being
            # joined and the merge_on columns that were present so far in the already
            # joined tables
            on = set(merge_on) & set(coalesced) & set(join.c.keys())

            # build the join condition
            cond = (
                and_(*(coalesced[c] == join.c[c] for c in on))
                if len(on) > 0
                else true()
                # true() is when there's no merge_on
                # or no common columns just cross join
            )

            # add the new join
            joined = joined.outerjoin(join, cond, full=True)

            # update the coalesced column list
            for column in set(merge_on) & set(join.c.keys()):
                if column in coalesced:
                    coalesced[column] = coalesce(coalesced[column], join.c[column])
                else:
                    coalesced[column] = join.c[column]

        # at this point we just need to select the coalesced columns
        columns = []
        for k, v in coalesced.items():
            columns.append(v.label(k))

        # and any other columns that are not in the coalesced mapping
        for s in subqueries:
            columns.extend(c for c in s.c if c.name not in coalesced)

        return select(*columns).select_from(joined)

    def calculate(self, query: Select, columns: List[Column]) -> Select:
        result_columns = []
        subquery = query.subquery()
        transformer = ColumnTransformer({None: subquery})
        for column in columns:
            resolved_column = transformer.transform(column.expr)
            sql_expr = self.transformer.transform(resolved_column).label(column.name)
            result_columns.append(sql_expr)
        return select(*result_columns).select_from(subquery)

    def inner_join(self, query: Select, to_join: Select, join_on: List[str]):
        to_join = to_join.subquery()
        conditions = and_(
            *(
                query.selected_columns[col] == to_join.c[col]
                for col in join_on
                if col in to_join.c  # support uneven level of detail
            )
        )
        return query.join(to_join, conditions)

    def limit(self, query: Select, limit: int):
        return query.limit(limit)

    def order(self, query: Select, items: List[LiteralOrderItem]) -> Select:
        clauses = []
        for item in items:
            clause = query.selected_columns[item.name]
            clause = clause.asc() if item.ascending else clause.desc()
            clauses.append(clause)
        return query.order_by(*clauses)

    def filter(self, query: Select, conditions: List[Tree]) -> Select:
        query = query.subquery().select()
        column_transformer = ColumnTransformer({None: query})
        for expr in conditions:
            condition = self.transformer.transform(column_transformer.transform(expr))
            query = query.where(condition)
        return query

    def filter_with_records(self, query: Select, records: List[List[Dict[str, Any]]]):
        if len(records) == 0:
            return query

        for recordset in records:
            conditions = []
            for record in recordset:
                # some filter columns might be missing from the query
                # for example if the query calculated a total
                condition_list = [
                    query.selected_columns[k] == v
                    for k, v in record.items()
                    if k in query.selected_columns
                ]
                if len(condition_list) > 0:
                    conditions.append(and_(*condition_list))
            if len(conditions) > 0:
                query = query.where(or_(*conditions))

        return query


class SQLAlchemyBackend(Backend):

    compiler_cls = SQLAlchemyCompiler

    def __init__(
        self,
        pool_size: Optional[int] = None,
        default_schema: Optional[str] = None,
        **kwargs,
    ):
        self.parameters = kwargs
        self.pool_size = pool_size
        self.default_schema = default_schema
        super().__init__()

    @property
    def url(self) -> str:
        url_params = {
            k: v
            for k, v in self.parameters.items()
            if k not in {"default_schema", "pool_size"}
        }
        return URL.create(**url_params)

    @cached_property
    def engine(self):
        return create_engine(self.url, pool_size=self.pool_size)

    @cached_property
    def metadata(self) -> MetaData:
        return MetaData(self.engine)

    def __str__(self):
        return repr(self.engine.url)

    def display_query(self, query: Select) -> str:
        return sqlparse.format(
            str(query.compile(bind=self.engine)),
            reindent=True,
        )

    def execute(self, query: Select) -> DataFrame:
        return read_sql(query, self.engine, coerce_float=True)

    def table(self, name: str, schema: Optional[str] = None) -> Table:
        if schema is None:
            schema = self.default_schema
        return Table(name, self.metadata, schema=schema, autoload=True)



from collections import defaultdict
from dataclasses import dataclass, field
from functools import cached_property
from typing import Dict, List, Optional, Tuple

from lark import Tree

from dictum_core.model.calculations import Dimension, TableFilter
from dictum_core.model.dicts import DimensionDict, MeasureDict


@dataclass
class RelatedTable:
    str_table: str
    foreign_key: str
    str_related_key: str
    alias: str

    parent: "Table"
    tables: Dict[str, "Table"]

    @property
    def table(self) -> "Table":
        if self.str_table not in self.tables:
            raise KeyError(
                f"Table {self.str_table} is referenced as related to {self.parent.id} "
                f"with alias {self.alias}, but it doesn't exist in the config"
            )
        return self.tables[self.str_table]

    @cached_property
    def related_key(self) -> str:
        if self.str_related_key is not None:
            return self.str_related_key
        if self.table.primary_key is None:
            raise ValueError(
                f"Table {self.table.id} with alias {self.alias} "
                f"is referenced as related to {self.parent.id}, "
                f"but neither primary_key on {self.table.id} nor "
                "related_key on the relationship are specified."
            )
        return self.table.primary_key

    @cached_property
    def join_expr(self) -> Tree:
        return Tree(
            "expr",
            [
                Tree(
                    "eq",
                    [
                        Tree("column", [self.parent.id, self.foreign_key]),
                        Tree("column", [self.parent.id, self.alias, self.related_key]),
                    ],
                )
            ],
        )


@dataclass(repr=False)
class Table:
    """Represents a relational data table"""

    id: str
    source: str
    description: Optional[str] = None
    primary_key: Optional[str] = None
    filters: List[TableFilter] = field(default_factory=list)
    related: Dict[str, RelatedTable] = field(default_factory=dict)
    measures: MeasureDict = field(default_factory=MeasureDict)
    dimensions: DimensionDict = field(default_factory=DimensionDict)
    measure_backlinks: Dict[str, "Table"] = field(
        default_factory=dict
    )  # measure_id -> table

    def find_all_paths(
        self, traversed_tables: Tuple[str] = ()
    ) -> List[Tuple["Table", List["str"]]]:
        """Find all join paths from this table to other tables, avoiding cycles"""
        traversed_tables += (self.id,)
        for rel in self.related.values():
            if isinstance(rel.table, Table) and rel.table.id not in traversed_tables:
                yield rel.table, [rel.alias]
                for target, path in rel.table.find_all_paths(traversed_tables):
                    yield target, [rel.alias, *path]

    @cached_property
    def allowed_join_paths(self) -> Dict["Table", List[str]]:
        """A dict of table id -> tuple list of related table aliases. Join targets for
        which there exists only a single join path.
        """
        paths = defaultdict(lambda: [])
        for target, path in self.find_all_paths():
            paths[target].append(path)
        result = {t: v[0] for t, v in paths.items() if len(v) == 1}
        for rel in self.related.values():
            result[rel.table] = [rel.alias]
        return result

    @cached_property
    def dimension_join_paths(self) -> Dict[str, List[str]]:
        result = {}
        for target, path in self.allowed_join_paths.items():
            for key, dim in target.dimensions.items():
                if not dim.is_union:
                    result[key] = [self.id] + path
        for dimension_id in self.dimensions:
            result[dimension_id] = [self.id]
        return result

    @cached_property
    def allowed_dimensions(self) -> Dict[str, "Dimension"]:
        """Which dimensions are allowed to be used with this table as anchor.
        Only those to which there's a single direct join path. If there isn't,
        dimensions must be declared directly on a table that's available for join.
        Unions are not allowed for joins.
        """
        result = {}
        for target in self.allowed_join_paths.keys():
            result.update(
                {d.id: d for d in target.dimensions.values() if not d.is_union}
            )
        result.update(self.dimensions)
        return result

    def __eq__(self, other: "Table"):
        return isinstance(other, Table) and self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"Table({self.id})"



from dictum_core.utils.version import find_latest_compatible_version


def test_find_version():
    assert (
        find_latest_compatible_version(
            "0.1.0",
            ["0.0\n", "0.1.0-dev123\n", "0.1.2-dev3\n", "0.1.14-dev1\n", "0.4.18\n"],
        )
        == "0.1.14-dev1"
    )



import datetime
from typing import List

import pandas as pd
from altair.utils.data import to_values

import dictum_core.project
import dictum_core.project.calculations
from dictum_core import engine, schema
from dictum_core.schema import Query


class DictumData:
    def __init__(
        self,
        project: "dictum_core.project.Project",
        metrics: List["dictum_core.project.calculations.ProjectMetric"],
    ):
        self.project = project
        self.requests = [m.request for m in metrics]
        self.filters = []
        self.limits = []

    def extend_query(self, query: Query):
        query = query.copy()
        for req in self.requests:
            if req not in query.metrics:
                query.metrics.append(req)
        query.filters.extend(self.filters)
        query.limit.extend(self.limits)
        return query

    def get_values(self, query):
        query = self.extend_query(query)
        res = self.project.execute(query)
        df = pd.DataFrame(res.data)

        # dates are not auto-converted to Pandas datetime and so are not sanitized
        for col in df.columns:
            if df[col].apply(lambda x: isinstance(x, datetime.date)).any():
                df[col] = pd.to_datetime(df[col])

        return to_values(df)

    def execute(self, query: schema.Query) -> engine.Result:
        query = self.extend_query(query)
        return self.project.execute(query)

    def to_dict(self):
        return {"name": "__dictum__"}



from typing import Dict, Optional

from pydantic import BaseModel, root_validator, validator

from dictum_core.schema import utils
from dictum_core.schema.id import ID
from dictum_core.schema.model.calculations import (
    DetachedDimension,
    DimensionsUnion,
    Metric,
)
from dictum_core.schema.model.table import Table
from dictum_core.schema.model.transform import Transform

root_keys = {"tables", "metrics", "unions"}


class Model(BaseModel):
    name: str
    description: Optional[str]
    locale: str = "en_US"
    currency: str = "USD"

    dimensions: Dict[ID, DetachedDimension] = {}
    metrics: Dict[ID, Metric] = {}
    unions: Dict[ID, DimensionsUnion] = {}

    tables: Dict[ID, Table] = {}
    transforms: Dict[
        ID, Transform
    ] = {}  # ignored for now, TODO: load as LiteralTransform

    theme: Optional[dict]

    set_metrics_ids = validator("metrics", allow_reuse=True, pre=True)(utils.set_ids)
    set_dimensions_ids = validator("dimensions", allow_reuse=True, pre=True)(
        utils.set_ids
    )
    set_tables_ids = validator("tables", allow_reuse=True, pre=True)(utils.set_ids)
    set_unions_ids = validator("unions", allow_reuse=True, pre=True)(utils.set_ids)
    set_transforms_ids = validator("transforms", allow_reuse=True, pre=True)(
        utils.set_ids
    )

    @root_validator(pre=True)
    def set_default_currency(cls, value: dict):
        currency = value.get("currency", "USD")

        def _set(calculation: dict):
            if calculation.get("format") == "currency":
                calculation["format"] = dict(kind="currency", currency=currency)
            elif isinstance(calculation.get("format"), str):
                return
            elif calculation.get("format", {}).get("kind") == "currency":
                calculation["format"].setdefault("currency", currency)

        for table in value.get("tables", {}).values():
            for key in ["measures", "dimensions"]:
                for calculation in table.get(key, {}).values():
                    _set(calculation)
        for key in ["metrics", "dimensions", "unions"]:
            for calculation in value.get(key, {}).values():
                _set(calculation)
        return value



import shutil
from pathlib import Path

import typer
import yaml
from jinja2 import Template
from rich.console import Console
from rich.prompt import IntPrompt, Prompt

from dictum_core.backends.base import Backend

app = typer.Typer()
console = Console()

backends = list(Backend.registry)

template_path = Path(__file__).parent / "project_template"


@app.command()
def new(
    project_dir: str = typer.Argument(None, help="Project directory"),
    project_name: str = typer.Option(None, help="Project name"),
    backend: str = typer.Option(None, help="Project backend"),
    default_profile: str = typer.Option(None, help="Default profile name"),
):
    if project_dir is None:
        project_dir = Prompt.ask("[bold blue]Project directory")
    project_dir = Path(project_dir)
    if project_dir.is_file():
        console.print("[red]Project path is a file")
        exit()
    if project_dir.is_dir() and not next(project_dir.iterdir(), None) is None:
        console.print("[red]Project directory must be empty")
        exit()
    if project_name is None:
        project_name = Prompt.ask("[bold blue]Project name", default=project_dir.name)
    if backend is None:
        backend_options = "\n".join(f"{i+1}. {name}" for i, name in enumerate(backends))
        backend_prompt = (
            f"[bold blue]Select a backend for your project:[/]\n\n{backend_options}\n"
            f"\nEnter a number [1-{len(backends)}]"
        )
        n = IntPrompt.ask(backend_prompt)
        backend = backends[n - 1]
    if default_profile is None:
        default_profile = Prompt.ask(
            "Enter a profile name for this backend", default=backend
        )

    if not project_dir.exists():
        project_dir.mkdir()

    backend_parameters = yaml.safe_dump(
        Backend.registry[backend].parameters(), sort_keys=False
    )
    template_vars = {
        "project_name": project_name,
        "backend": backend,
        "profile": default_profile,
        "backend_parameters": backend_parameters,
    }

    for path in template_path.iterdir():
        if path.name in {"__init__.py", ".gitkeep"}:
            continue
        new_path = project_dir / path.relative_to(template_path)
        if path.is_file():
            template = Template(path.read_text())
            rendered = template.render(**template_vars)
            new_path.write_text(rendered)
        elif path.is_dir():
            shutil.copytree(str(path), str(new_path))
    console.print(f"[bold green]Project {project_name} created")


@app.command()
def version():
    from dictum_core import __version__

    typer.echo(f"Dictum {__version__}")



import inspect
from collections import UserDict
from functools import cached_property
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import pkg_resources
from lark import Token, Transformer
from pandas import DataFrame

from dictum_core.engine import Column, LiteralOrderItem, RelationalQuery


@dataclass
class BackendResult:
    data: List[dict]
    duration: float  # ms
    raw_query: str


class CallOwnerOp:
    def __get__(self, owner, ownertype=None):
        self._owner = owner
        return self

    def __set_name__(self, owner: "Transformer", name: str):
        self._owner = owner
        self._name = name

    def __call__(self, children: list):
        raise NotImplementedError


class PassChildrenToCompiler(CallOwnerOp):
    def __call__(self, children: list):
        return getattr(self._owner.compiler, self._name)(*children)


class PassTokenValueToCompiler(CallOwnerOp):
    def __call__(self, token: Token):
        return getattr(self._owner.compiler, self._name)(token.value)


def _get_whens_else(args: list):
    whens = [args[i : i + 2] for i in range(0, len(args), 2)]  # noqa: E203
    else_ = None if len(whens[-1]) == 2 else whens.pop(-1)[0]
    return whens, else_


class ExpressionTransformer(Transformer):
    def __init__(self, compiler: "Compiler", visit_tokens: bool = True) -> None:
        self.compiler = compiler
        super().__init__(visit_tokens=visit_tokens)

    def ARG(self, _):
        raise ValueError(
            "'@' is not allowed in calculation expressions —"
            " only in user-defined filters and transforms"
        )

    FLOAT = PassTokenValueToCompiler()
    INTEGER = PassTokenValueToCompiler()
    STRING = PassTokenValueToCompiler()

    def TRUE(self, _):
        return self.compiler.TRUE()

    def FALSE(self, _):
        return self.compiler.FALSE()

    column = PassChildrenToCompiler()

    exp = PassChildrenToCompiler()
    neg = PassChildrenToCompiler()
    fdiv = PassChildrenToCompiler()
    div = PassChildrenToCompiler()
    mul = PassChildrenToCompiler()
    mod = PassChildrenToCompiler()
    add = PassChildrenToCompiler()
    sub = PassChildrenToCompiler()
    gt = PassChildrenToCompiler()
    ge = PassChildrenToCompiler()
    lt = PassChildrenToCompiler()
    le = PassChildrenToCompiler()
    eq = PassChildrenToCompiler()
    ne = PassChildrenToCompiler()

    isnull = PassChildrenToCompiler()

    def IN(self, children: list):
        value, *values = children
        return self.compiler.IN(value, values)

    NOT = PassChildrenToCompiler()
    AND = PassChildrenToCompiler()
    OR = PassChildrenToCompiler()

    def case(self, children: list):
        whens, else_ = _get_whens_else(children)
        return self.compiler.case(whens, else_=else_)

    def expr(self, children: list):
        return children[0]

    def call(self, children: list):
        fn, *args = children
        return self.compiler.call(fn, args)

    def order_by(self, children: list):
        return children

    def call_window(self, children: list):
        fn, *args, partition, order, rows = children
        return self.compiler.call_window(fn, args, partition, order, rows)

    def partition_by(self, children: list):
        return children


class Compiler(ABC):
    """Takes in a computation, returns an object that a connection will understand."""

    def __init__(self):
        self.transformer = ExpressionTransformer(self)

    # expression language elements

    @abstractmethod
    def FLOAT(self, value: str):
        """Float literal"""

    @abstractmethod
    def INTEGER(self, value: str):
        """Integer literal"""

    @abstractmethod
    def STRING(self, value: str):
        """String literal"""

    @abstractmethod
    def TRUE(self):
        """True boolean literal"""

    @abstractmethod
    def FALSE(self):
        """False boolean literal"""

    @abstractmethod
    def isnull(self, value):
        """Missing value check"""

    @abstractmethod
    def column(self, table: str, name: str):
        """Column reference.
        table is a dot-delimited sequence of identifiers
        name is the column name
        """

    def call(self, fn: str, args: list):
        """Function call. First element of children is function name, rest are
        the arguments. Calls the method on self.
        """
        call = getattr(self, fn)
        return call(*args)

    def call_window(
        self, fn: str, args: list, partition: list, order: list, rows: list
    ):
        fn = getattr(self, f"window_{fn}")
        return fn(args, partition, order, rows)

    @abstractmethod
    def exp(self, a, b):
        """Exponentiation — "power" operator, a to the power of b"""

    @abstractmethod
    def neg(self, value):
        """Unary number negation, e.g. -1"""

    @abstractmethod
    def div(self, a, b):
        """Normal division. Semantics depend on the underlying backend."""

    @abstractmethod
    def mul(sef, a, b):
        """Arithmetic multiplication"""

    @abstractmethod
    def mod(self, a, b):
        """Modulo, arithmetic remainder, e.g. 7 % 2 == 1"""

    @abstractmethod
    def add(self, a, b):
        """Arithmetic addition"""

    @abstractmethod
    def sub(self, a, b):
        """Arithmetic subtraction"""

    @abstractmethod
    def gt(self, a, b):
        """Greater than, a > b"""

    @abstractmethod
    def ge(self, a, b):
        """Greater than or equal, a >= b"""

    @abstractmethod
    def lt(self, a, b):
        """Less than, a < b"""

    @abstractmethod
    def le(self, a, b):
        """Less than or equal, a <= b"""

    @abstractmethod
    def eq(self, a, b):
        """Equality, a equals b"""

    @abstractmethod
    def ne(self, a, b):
        """Non-equality, a not equals b"""

    @abstractmethod
    def IN(self, value, values):
        """Value is in a tuple of values"""

    @abstractmethod
    def NOT(self, value):
        """Boolean negation, NOT x"""

    @abstractmethod
    def AND(self, a, b):
        """Logical conjunction"""

    @abstractmethod
    def OR(self, a, b):
        """Logical disjunction"""

    @abstractmethod
    def case(self, whens, else_=None):
        """whens: tuples of (condition, value)
        else: else value (optional)
        """

    def IF(self, *args):
        whens, else_ = _get_whens_else(args)
        return self.case(whens, else_=else_)

    # built-in functions
    # aggregate

    @abstractmethod
    def sum(self, arg):
        """Aggregate sum"""

    @abstractmethod
    def avg(self, arg):
        """Aggregate average"""

    @abstractmethod
    def min(self, arg):
        """Aggregate minimum"""

    @abstractmethod
    def max(self, arg):
        """Aggregate maximum"""

    @abstractmethod
    def count(self, arg=None):  # arg can be missing
        """Aggregate count, with optional argument"""

    @abstractmethod
    def countd(self, arg):
        """Aggregate distinct count"""

    # window functions

    @abstractmethod
    def window_sum(self, arg, partition_by, order_by, rows):
        """A windowed version of aggregate sum function"""

    @abstractmethod
    def window_row_number(self, arg, partition_by, order_by, rows):
        """Same as SQL row_number"""

    def window_avg(self, arg, partition, order, rows):
        """TODO"""

    def window_min(self, arg, partition, order, rows):
        """TODO"""

    def window_max(self, arg, partition, order, rows):
        """TODO"""

    def window_count(self, arg, partition, order, rows):
        """TODO"""

    # scalar functions

    @abstractmethod
    def abs(self, arg):
        """Absolute numeric value"""

    @abstractmethod
    def floor(self, arg):
        """Numeric floor"""

    @abstractmethod
    def ceil(self, arg):
        """Numeric ceiling"""

    @abstractmethod
    def coalesce(self, *args):
        """NULL-coalescing"""

    # type casting

    @abstractmethod
    def tointeger(self, arg):
        """cast as int"""

    @abstractmethod
    def tofloat(self, arg):
        """cast as float"""

    @abstractmethod
    def todate(self, arg):
        """cast as date"""

    @abstractmethod
    def todatetime(self, arg):
        """cast as datetime/timestamp"""

    # dates

    @abstractmethod
    def datepart(self, part, arg):
        """Part of a date as an integer. First arg is part as a string, e.g. 'month',
        second is date/datetime.
        """

    @abstractmethod
    def datetrunc(self, part, arg):
        """Date truncated to a given part. Args same as datepart."""

    @abstractmethod
    def datediff(self, part, start, end):
        """Difference between two dates, given as number of times there's a change of
        date at the given level.
        """

    @abstractmethod
    def now(self):
        """Current timestamp"""

    @abstractmethod
    def today(self):
        """Today's date"""

    # compilation

    @abstractmethod
    def compile_query(self, query: RelationalQuery):
        """Compile a single relational query into connection query."""

    @abstractmethod
    def merge_queries(self, queries: List, merge_on: List[str]):
        """Merge a list of relational queries on the relevant level of detail. Can be
        unsupported by a backend, raise NotImplementedError in this case.
        """

    @abstractmethod
    def calculate(self, query, columns: List[Column]):
        """Calculate expressions from the fields of a query"""

    @abstractmethod
    def filter(self, query, conditions: Dict[str, Any]):
        """Filter a query on a list of conditions"""

    @abstractmethod
    def filter_with_records(self, query, records: List[List[Dict[str, Any]]]):
        """Use a list of dicts as a filter for a query. Each dict is a
        combination of valid field values, the rest are filtered out.
        """

    @abstractmethod
    def inner_join(self, query, to_join, join_on: List[str]):
        """Use another query as an inner join for a query, join condition is equality
        on a list of columns (same names in both queries).
        """

    @abstractmethod
    def limit(self, query, limit: int):
        """Limit a query (mostly user with order)."""

    @abstractmethod
    def order(self, query, items: List[LiteralOrderItem]):
        """Add ordering to a query."""


class BackendRegistry(UserDict):
    def __init__(self, dict=None, /, **kwargs):
        super().__init__(dict, **kwargs)

    @cached_property
    def registry(self):
        Backend.discover_plugins()
        return self.data

    def __getitem__(self, key: str) -> "Backend":
        if key not in self.registry:
            raise ImportError(
                f"Backend {type} was not found. Try installing dictum[{type}] "
                "package."
            )
        return self.data[key]

    def __contains__(self, key: object) -> bool:
        return key in self.registry

    def __str__(self) -> str:
        return str(self.registry)

    def __iter__(self):
        return iter(self.registry)


class Backend(ABC):
    """User-facing. Gets connection details, knows about it's compiler. Compiles
    the incoming computation, executes on the client.
    """

    type: str
    compiler_cls = Compiler

    registry: BackendRegistry = BackendRegistry()

    def __init__(self):
        self.compiler = self.compiler_cls(self)

    def __init_subclass__(cls):
        if hasattr(cls, "type"):
            cls.registry[cls.type] = cls

    @classmethod
    def create(cls, type: str, parameters: Optional[dict] = None):
        if parameters is None:
            parameters = {}
        return cls.registry[type](**parameters)

    @classmethod
    def parameters(cls) -> dict:
        result = {}
        for name, par in inspect.signature(cls.__init__).parameters.items():
            if name == "self":
                continue
            result[name] = par.default if par.default != inspect._empty else None
        return result

    @classmethod
    def discover_plugins(cls):
        for entry_point in pkg_resources.iter_entry_points("dictum.backends"):
            cls.registry[entry_point.name] = entry_point.load()

    def display_query(self, query):
        return str(query)

    def compile_query(self, query):
        return self.compiler.compile_query(query)

    def merge_queries(self, queries: List, merge_on: List[str]):
        return self.compiler.merge_queries(queries, merge_on)

    def calculate(self, query, columns: List[Column]):
        return self.compiler.calculate(query, columns)

    def filter(self, query, conditions: Dict[str, Any]):
        return self.compiler.filter(query, conditions)

    def filter_with_records(self, query, records: List[List[Dict[str, Any]]]):
        return self.compiler.filter_with_records(query, records)

    def inner_join(self, query, to_join, join_on: List[str]):
        return self.compiler.inner_join(query, to_join, join_on)

    def limit(self, query, limit: int):
        return self.compiler.limit(query, limit)

    def order(self, query, items: List[LiteralOrderItem]):
        return self.compiler.order(query, items)

    @abstractmethod
    def execute(self, query) -> DataFrame:
        """Execute query, return results"""



from functools import lru_cache
from typing import Optional

from babel.core import Locale

import dictum_core.project.altair.format

currency_symbol_overrides = {
    "RUB": "₽",
    None: "",
}

cache = lru_cache(maxsize=None)


@cache
def load_locale(locale: str) -> Locale:
    return Locale(locale)


def cldr_locale_to_d3_number(locale: str, currency: Optional[str] = None):
    data = load_locale(locale)
    fix, start = "", True
    if currency is not None:
        symbol = currency_symbol_overrides.get(currency)
        if symbol is None:
            symbol = data.currency_symbols.get(currency, currency)
        fmt = data.currency_formats["standard"]
        start = fmt.pattern.startswith("¤")
        fix = fmt.prefix[0] if start else fmt.suffix[0]
        fix = fix.replace("¤", symbol)
    return {
        "decimal": data.number_symbols["decimal"],
        "thousands": data.number_symbols["group"],
        "grouping": list(next(iter(data.decimal_formats.values())).grouping),
        "currency": [fix, ""] if start else ["", fix],
    }


def cldr_locale_to_d3_time(locale: str):
    data = load_locale(locale)
    days = [6] + list(range(6))  # days in d3 start from Sunday
    short_date = dictum_core.project.altair.format.ldml_date_to_d3_time_format(
        data.date_formats["short"].pattern
    )
    short_time = dictum_core.project.altair.format.ldml_date_to_d3_time_format(
        data.time_formats["short"].pattern
    )
    datetime = data.datetime_formats["short"].format(short_time, short_date)
    return {
        "dateTime": datetime,
        "date": short_date,
        "time": short_time,
        "periods": [
            data.day_periods["format"]["abbreviated"]["am"],
            data.day_periods["format"]["abbreviated"]["pm"],
        ],
        "days": [data.days["format"]["wide"][k] for k in days],
        "shortDays": [data.days["format"]["abbreviated"][k] for k in days],
        "months": [data.months["format"]["wide"][k] for k in range(1, 13)],
        "shortMonths": [data.months["format"]["abbreviated"][k] for k in range(1, 13)],
    }


def cldr_locale_to_d3(locale: str, currency: Optional[str] = None):
    return {
        "number": cldr_locale_to_d3_number(locale, currency=currency),
        "time": cldr_locale_to_d3_time(locale),
    }



from typing import List, Optional
import sys
from packaging import version
from dictum_core import __version__


def find_latest_compatible_version(target: str, candidates: List[str]) -> Optional[str]:
    target = version.parse(target)
    result = None

    for current in candidates:
        current = current.strip()
        ver = version.parse(current)
        if ver.major == target.major and ver.minor == target.minor:
            if result is None or version.parse(result) < ver:
                result = current

    return result


if __name__ == "__main__":
    target = __version__
    if len(sys.argv) > 1:
        target = sys.argv[1]
    result = find_latest_compatible_version(target, sys.stdin)
    print(result)



import os
from pathlib import Path

import pytest

from dictum_core import schema
from dictum_core.examples import chinook
from dictum_core.model import Model

chinook_path = Path(chinook.__file__).parent


@pytest.fixture(scope="session")
def backend():
    from dictum_core.examples.chinook.generate import generate

    yield generate().backend


@pytest.fixture(scope="session")
def project(backend):
    from dictum_core import Project

    project = Project.example("chinook")
    project.backend = backend
    yield project


@pytest.fixture(scope="session")
def chinook():
    os.environ["CHINOOK_DATABASE"] = ""
    return Model(schema.Project.load(chinook_path).get_model())


@pytest.fixture(scope="session")
def engine(chinook):
    from dictum_core.engine import Engine

    return Engine(chinook)



"""Mostly tests that Altair integration works without errors. Tests must return the
chart, which will be rendered and saved to altair_output folder (see conftest.py).

When possible, the tests check that the output is correct, but generally it's better
to visually inspect the output.
"""

import altair as alt

from dictum_core import Project


def test_metric(project: Project):
    return project.chart().mark_bar().encode(x=project.m.revenue)


def test_metric_dimension(project: Project):
    return (
        project.chart()
        .mark_bar()
        .encode(
            x=project.d.invoice_date.year.name("Year"),
            y=project.m.revenue,
        )
    )


def test_metric_dimension_year(project: Project):
    return (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.d.Year, type="ordinal"),
            y=project.m.revenue,
        )
    )


def test_override_type(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.d.invoice_date.year.name("Year"), type="ordinal"),
            y=alt.Y(project.m.revenue),
        )
    )
    assert chart._render_self().encoding.x.type == "ordinal"
    return chart


def test_override_axis_title(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(x=alt.X(project.m.revenue, axis=alt.Axis(title="Revenue ($)")))
    )
    assert chart._render_self().encoding.x.axis["title"] == "Revenue ($)"
    return chart


def test_inject_format(project: Project):
    chart = project.chart().mark_bar().encode(x=project.m.revenue)
    assert chart._render_self().encoding.x.axis["format"] == "$01,.2f"
    return chart


def test_facet(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.d.invoice_date.year, type="ordinal"),
            y=alt.Y(project.m.revenue),
        )
        .properties(width=150, height=100)
        .facet(project.d.genre, columns=3)
    )
    assert chart._render_self().facet["header"]["title"] == "Genre"
    return chart


def test_facet_row(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.d.invoice_date.year, type="ordinal"),
            y=alt.Y(project.m.revenue),
        )
        .properties(width=150, height=100)
        .facet(row=project.d.genre)
    )
    assert chart._render_self().facet.row["header"]["title"] == "Genre"
    return chart


def test_facet_column(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.d.invoice_date.year, type="ordinal"),
            y=alt.Y(project.m.revenue),
        )
        .properties(width=150, height=100)
        .facet(column=project.d.genre)
    )
    assert chart._render_self().facet.column["header"]["title"] == "Genre"
    return chart


def test_repeat(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.d.invoice_date.year, type="ordinal"),
            y=alt.Y(alt.repeat(), type="quantitative"),
        )
        .properties(width=150, height=100)
        .repeat([project.m.revenue, project.m.items_sold])
    )
    rendered = chart._rendered_dict()
    assert len(rendered["spec"]["encoding"]["tooltip"]) == 3
    return chart


def test_repeat_row(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.d.invoice_date.year, type="ordinal"),
            y=alt.Y(alt.repeat("row"), type="quantitative"),
        )
        .properties(width=150, height=100)
        .repeat(row=[project.m.revenue, project.m.items_sold])
    )
    return chart


def test_repeat_column(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.d.invoice_date.year, type="ordinal"),
            y=alt.Y(alt.repeat("column"), type="quantitative"),
        )
        .properties(width=150, height=100)
        .repeat(column=[project.m.revenue, project.m.items_sold])
    )
    return chart


def test_sort(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.m.revenue),
            y=alt.Y(project.d.genre, sort=project.m.revenue),
        )
    )
    return chart


def test_sort_field(project: Project):
    chart = (
        project.chart()
        .mark_bar()
        .encode(
            x=alt.X(project.m.revenue),
            y=alt.Y(
                project.d.genre,
                sort=alt.EncodingSortField(project.m.track_count, order="descending"),
            ),
        )
    )
    return chart


def test_detail(project: Project):
    chart = (
        project.chart()
        .mark_bar(stroke="white")
        .encode(
            x=alt.X(project.d.invoice_date.year, type="ordinal"),
            y=alt.Y(project.m.revenue),
            detail=project.d.genre,
        )
    )
    return chart


def test_detail_list(project: Project):
    chart = (
        project.chart()
        .mark_bar(stroke="white")
        .encode(
            x=alt.X(project.d.invoice_date.year, type="ordinal"),
            y=alt.Y(project.m.revenue),
            detail=[project.d.genre],
        )
    )
    return chart


def test_calculate_transform(project: Project):
    """Test that normal encodings derived from transform_calculate'd fields work as
    expected.
    """
    chart = (
        project.chart(project.m.revenue)
        .mark_bar()
        .encode(x=project.d.invoice_date.year, y="max_pct:Q")
        .transform_joinaggregate(
            groupby=[],
            max="sum(revenue)",
        )
        .transform_calculate(max_pct="datum.revenue / datum.max")
    )
    return chart


def test_custom_calculation_format(project: Project):
    return (
        project.chart()
        .mark_bar()
        .encode(
            x=project.m.revenue_per_track,
            y=project.d.genre,
        )
    )



from dictum_core.schema import FormatConfig


class TimeDimension:
    def __init__(cls, name, bases, attrs):
        cls.id = name
        cls.name = name
        cls.period = attrs.get("period")
        cls.pattern = attrs.get("pattern")
        cls.skeleton = attrs.get("skeleton")

    def __repr__(self):
        return self.name

    @property
    def format(self) -> FormatConfig:
        if self.skeleton:
            return FormatConfig(kind="datetime", skeleton=self.skeleton)
        if self.pattern:
            return FormatConfig(kind="datetime", pattern=self.pattern)
        raise ValueError  # this shouldn't happen


class BaseTimeDimension(metaclass=TimeDimension):
    pattern: str = None
    skeleton: str = None

    def __init__(self):
        raise ValueError("Time dimensions are singletons, don't instantiate them")


class Time(BaseTimeDimension):
    period = None
    skeleton = "yyMMdHmmss"


class Year(BaseTimeDimension):
    period = "year"
    pattern = "yy"


class Quarter(BaseTimeDimension):
    period = "quarter"
    pattern = "qqqq yy"


class Month(BaseTimeDimension):
    period = "month"
    skeleton = "MMMMy"


class Week(BaseTimeDimension):
    period = "week"
    skeleton = "w Y"


class Day(BaseTimeDimension):
    period = "day"
    skeleton = "yMd"


class Date(Day):
    period = "day"


class Hour(BaseTimeDimension):
    period = "hour"
    skeleton = "yMd hh"


class Minute(BaseTimeDimension):
    period = "minute"
    skeleton = "yMd hm"


class Second(BaseTimeDimension):
    period = "second"
    skeleton = "yMd hms"


dimensions = {
    "Time": Time,
    "Year": Year,
    "Quarter": Quarter,
    "Month": Month,
    "Week": Week,
    "Day": Day,
    "Date": Date,
    "Hour": Hour,
    "Minute": Minute,
    "Second": Second,
}



from dataclasses import dataclass
from typing import List

from toolz import compose_left

from dictum_core.engine.computation import Column, RelationalQuery
from dictum_core.engine.result import DisplayInfo
from dictum_core.model import DimensionsUnion, Measure, Model
from dictum_core.model.scalar import DatetruncTransform
from dictum_core.model.time import TimeDimension
from dictum_core.schema import QueryDimension, QueryDimensionRequest


@dataclass
class AggregateQueryBuilder:
    """Holds a list of dimensions and filters that must be applied
    for a given query.

    Contains helper methods that allow to build a query with necessary
    dimensions and filters from any given anchor table.
    """

    model: Model
    dimensions: List[QueryDimensionRequest]
    filters: List[QueryDimension]

    def get_dimension_column(
        self, measure: Measure, request: QueryDimensionRequest
    ) -> Column:
        """Construct a Column with a transformed expression relative to
        a given measure's anchor table.
        """
        anchor = measure.table
        dimension = self.model.dimensions.get(request.dimension.id)

        # if union, replace dimension with it
        if isinstance(dimension, DimensionsUnion):
            union = dimension
            dimension = anchor.allowed_dimensions.get(union.id)
            if dimension is None:
                raise ValueError(f"Can't use {union} with {measure}")

        transforms = [
            self.model.scalar_transforms[t.id](*t.args)
            for t in request.dimension.transforms
        ]

        display_name = request.alias if request.alias else dimension.name

        # if generic time, prepend transforms with datetrunc
        # and replace the dimension with measure's time
        if isinstance(dimension, TimeDimension):
            if dimension.period is not None:
                transforms = [DatetruncTransform(dimension.period), *transforms]
                if measure.time is None:
                    raise ValueError(
                        f"You requested a generic Time dimension with {measure}, "
                        "but it doesn't have a time dimension specified"
                    )
            display_name = dimension.name if request.alias is None else request.alias
            dimension = measure.time

        # get the expression with join info
        try:
            join_path = anchor.dimension_join_paths[dimension.id]
        except KeyError:
            raise KeyError(
                f"You requested {dimension}, but it can't be used with "
                f"another measure that you requested on {anchor}"
            )
        expr = dimension.prefixed_expr(join_path)

        result = Column(
            name=request.digest,
            expr=expr,
            type=dimension.type,
            display_info=DisplayInfo(
                display_name=display_name,
                column_name=request.name,
                format=dimension.format,
                kind="dimension",
                keep_display_name=(request.alias is not None),
            ),
        )

        return compose_left(*transforms)(result)

    def get_aggregation(self, measure: Measure) -> RelationalQuery:
        anchor = measure.table
        result = RelationalQuery(source=anchor, join_tree=[])

        for dimension in self.dimensions:
            column = self.get_dimension_column(measure, dimension)
            result.add_groupby(column)

        for filter in self.filters:
            # create a fake request because filters are not requests
            # and get_dimension_column expects a request
            request = QueryDimensionRequest(dimension=filter)
            column = self.get_dimension_column(measure, request)
            result.add_filter_expr(column.expr)  # add just the expr

        for filter in anchor.filters:
            result.add_filter_expr(filter.expr)

        if measure.filter:
            result.add_filter_expr(measure.filter)

        result.add_aggregate(
            column=Column(name=measure.id, expr=measure.expr, type=measure.type)
        )

        return result



from pathlib import Path

from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader

path = Path(__file__).parent
environment = Environment(loader=FileSystemLoader(searchpath=path))



from lark import Tree

from dictum_core.ql import compile_query
from dictum_core.ql.parser import (
    parse_dimension,
    parse_dimension_request,
    parse_metric_request,
    parse_ql,
)
from dictum_core.ql.transformer import (
    compile_dimension,
    compile_dimension_request,
    compile_metric_request,
)
from dictum_core.schema.query import (
    Query,
    QueryDimension,
    QueryDimensionRequest,
    QueryMetricRequest,
)


def test_single_metric():
    assert parse_ql("select revenue") == Tree(
        "query",
        [Tree("select", [Tree("metric_request", [Tree("metric", ["revenue"])])])],
    )


def test_metric_transform():
    assert parse_ql("select x.sum(1, 2, 3) of (x) within (y, z)") == Tree(
        "query",
        [
            Tree(
                "select",
                [
                    Tree(
                        "metric_request",
                        [
                            Tree(
                                "metric",
                                [
                                    "x",
                                    Tree(
                                        "table_transform",
                                        [
                                            "sum",
                                            1,
                                            2,
                                            3,
                                            Tree("of", [Tree("dimension", ["x"])]),
                                            Tree(
                                                "within",
                                                [
                                                    Tree("dimension", ["y"]),
                                                    Tree("dimension", ["z"]),
                                                ],
                                            ),
                                        ],
                                    ),
                                ],
                            )
                        ],
                    )
                ],
            )
        ],
    )


def test_multiple_metrics():
    assert parse_ql("select revenue, test") == Tree(
        "query",
        [
            Tree(
                "select",
                [
                    Tree("metric_request", [Tree("metric", ["revenue"])]),
                    Tree("metric_request", [Tree("metric", ["test"])]),
                ],
            ),
        ],
    )


def test_where_transform():
    assert parse_ql("select x where y.z(1)") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "where", [Tree("dimension", ["y", Tree("scalar_transform", ["z", 1])])]
            ),
        ],
    )


def test_where_transforms():
    assert parse_ql("select x where y.a(1).b('c').d = 0") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "where",
                [
                    Tree(
                        "dimension",
                        [
                            "y",
                            Tree("scalar_transform", ["a", 1]),
                            Tree("scalar_transform", ["b", "c"]),
                            Tree("scalar_transform", ["d"]),
                            Tree("scalar_transform", ["eq", 0]),
                        ],
                    )
                ],
            ),
        ],
    )


def test_where_gt():
    assert parse_ql("select x where y > 0") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "where", [Tree("dimension", ["y", Tree("scalar_transform", ["gt", 0])])]
            ),
        ],
    )


def test_where_ge():
    assert parse_ql("select x where y >= 0") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "where", [Tree("dimension", ["y", Tree("scalar_transform", ["ge", 0])])]
            ),
        ],
    )


def test_where_lt():
    assert parse_ql("select x where y < 0") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "where", [Tree("dimension", ["y", Tree("scalar_transform", ["lt", 0])])]
            ),
        ],
    )


def test_where_le():
    assert parse_ql("select x where y <= 0") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "where", [Tree("dimension", ["y", Tree("scalar_transform", ["le", 0])])]
            ),
        ],
    )


def test_where_eq():
    assert parse_ql("select x where y = 0") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "where", [Tree("dimension", ["y", Tree("scalar_transform", ["eq", 0])])]
            ),
        ],
    )


def test_where_ne():
    assert parse_ql("select x where y <> 0") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "where", [Tree("dimension", ["y", Tree("scalar_transform", ["ne", 0])])]
            ),
        ],
    )


def test_groupby():
    assert parse_ql("select x by y") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree("groupby", [Tree("dimension_request", [Tree("dimension", ["y"])])]),
        ],
    )


def test_groupby_transform():
    assert parse_ql("select x group by y.z(10)") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "groupby",
                [
                    Tree(
                        "dimension_request",
                        [Tree("dimension", ["y", Tree("scalar_transform", ["z", 10])])],
                    )
                ],
            ),
        ],
    )


def test_quoted_identifier():
    assert parse_ql('select "revenue something"').children[0] == Tree(
        "select", [Tree("metric_request", [Tree("metric", ["revenue something"])])]
    )


def test_dimension_alias():
    assert parse_ql("select metric by dim as dim1") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["metric"])])]),
            Tree(
                "groupby",
                [
                    Tree(
                        "dimension_request",
                        [Tree("dimension", ["dim"]), Tree("alias", ["dim1"])],
                    )
                ],
            ),
        ],
    )


def test_dimension_transform_alias():
    assert parse_ql("select metric by dim.test as dim1") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["metric"])])]),
            Tree(
                "groupby",
                [
                    Tree(
                        "dimension_request",
                        [
                            Tree(
                                "dimension", ["dim", Tree("scalar_transform", ["test"])]
                            ),
                            Tree("alias", ["dim1"]),
                        ],
                    )
                ],
            ),
        ],
    )


def test_in():
    assert parse_ql("select x where y in ('a', 1)") == Tree(
        "query",
        [
            Tree("select", [Tree("metric_request", [Tree("metric", ["x"])])]),
            Tree(
                "where",
                [Tree("dimension", ["y", Tree("scalar_transform", ["isin", "a", 1])])],
            ),
        ],
    )


def test_compile_query():
    assert compile_query("select x") == Query.parse_obj(
        {"metrics": [{"metric": {"id": "x"}}]}
    )


def test_compile_groupby():
    q = Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "x"}}],
            "dimensions": [{"dimension": {"id": "y"}}],
        }
    )
    assert compile_query("select x group by y") == q
    assert compile_query("select x by y") == q


def test_compile_groupby_transforms():
    assert compile_query("select x, y by z.p(10)") == Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "x"}}, {"metric": {"id": "y"}}],
            "dimensions": [
                {"dimension": {"id": "z", "transforms": [{"id": "p", "args": [10]}]}}
            ],
        }
    )


def test_compile_where():
    assert compile_query("select x where y.z(10)") == Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "x"}}],
            "filters": [{"id": "y", "transforms": [{"id": "z", "args": [10]}]}],
        }
    )


def test_compile_multiple_groupbys():
    compile_query(
        """
    select x, y
    where z.z(1, 2, 3),
        a.b('a')
    by d.d('d'),
       c,
       f.h(11)
    """
    ) == Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "x"}}, {"metric": {"id": "y"}}],
            "dimensions": [
                {
                    "dimension": {
                        "id": "d",
                        "transforms": [{"id": "d", "args": ["d"]}],
                    },
                },
                {"dimension": {"id": "c"}},
                {"dimension": {"id": "f", "tranforms": [{"id": "h", "args": [11]}]}},
            ],
            "filters": [
                {
                    "id": "z",
                    "transforms": [{"id": "z", "args": [1, 2, 3]}],
                },
                {"id": "a", "transforms": [{"id": "b", "args": ["a"]}]},
            ],
        }
    )


def test_compile_dimension_alias():
    assert compile_query("select metric by dim as dim1") == Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "metric"}}],
            "dimensions": [{"dimension": {"id": "dim"}, "alias": "dim1"}],
        }
    )


def test_compile_dimension_transform_alias():
    assert compile_query("select x by y.a.b(10) as z") == Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "x"}}],
            "dimensions": [
                {
                    "dimension": {
                        "id": "y",
                        "transforms": [{"id": "a"}, {"id": "b", "args": [10]}],
                    },
                    "alias": "z",
                }
            ],
        }
    )


def test_parse_filter():
    assert parse_dimension("x.y('z')") == Tree(
        "dimension", ["x", Tree("scalar_transform", ["y", "z"])]
    )


def test_compile_filter():
    assert compile_dimension("x.y('z')") == QueryDimension.parse_obj(
        {"id": "x", "transforms": [{"id": "y", "args": ["z"]}]}
    )


def test_filter_null():
    assert parse_dimension("x is null") == Tree(
        "dimension", ["x", Tree("scalar_transform", ["isnull"])]
    )
    assert parse_dimension("x is not null") == Tree(
        "dimension", ["x", Tree("scalar_transform", ["isnotnull"])]
    )


def test_compile_filter_null():
    assert compile_dimension("x is null") == QueryDimension.parse_obj(
        {"id": "x", "transforms": [{"id": "isnull", "args": []}]}
    )


def test_parse_dimension_request():
    assert parse_dimension_request("x.y(1)") == Tree(
        "dimension_request",
        [Tree("dimension", ["x", Tree("scalar_transform", ["y", 1])])],
    )
    assert parse_dimension_request("x") == Tree(
        "dimension_request", [Tree("dimension", ["x"])]
    )


def test_compile_dimension_request():
    assert compile_dimension_request("x") == QueryDimensionRequest.parse_obj(
        {"dimension": {"id": "x"}}
    )
    assert compile_dimension_request("x.y") == QueryDimensionRequest.parse_obj(
        {"dimension": {"id": "x", "transforms": [{"id": "y"}]}}
    )
    assert compile_dimension_request(
        "x.y('z', 1) as a"
    ) == QueryDimensionRequest.parse_obj(
        {
            "dimension": {"id": "x", "transforms": [{"id": "y", "args": ["z", 1]}]},
            "alias": "a",
        }
    )


def test_parse_metric_request():
    assert parse_metric_request("x.y(1, 'f') of (a) within (b, c) as al") == Tree(
        "metric_request",
        [
            Tree(
                "metric",
                [
                    "x",
                    Tree(
                        "table_transform",
                        [
                            "y",
                            1,
                            "f",
                            Tree("of", [Tree("dimension", ["a"])]),
                            Tree(
                                "within",
                                [Tree("dimension", ["b"]), Tree("dimension", ["c"])],
                            ),
                        ],
                    ),
                ],
            ),
            Tree("alias", ["al"]),
        ],
    )


def test_compile_metric_request():
    assert compile_metric_request(
        "x.y(1, 'f') of (a) within (b, c.n(1)) as al"
    ) == QueryMetricRequest.parse_obj(
        {
            "metric": {
                "id": "x",
                "transforms": [
                    {
                        "id": "y",
                        "args": [1, "f"],
                        "of": [{"id": "a"}],
                        "within": [
                            {"id": "b"},
                            {"id": "c", "transforms": [{"id": "n", "args": [1]}]},
                        ],
                    }
                ],
            },
            "alias": "al",
        }
    )


def test_parse_unary_not():
    assert parse_dimension("x.y(1).not") == Tree(
        "dimension",
        [
            "x",
            Tree("scalar_transform", ["y", 1]),
            Tree("scalar_transform", ["invert"]),
        ],
    )



import dataclasses
from typing import Dict

from dictum_core import schema
from dictum_core.model.calculations import Dimension, DimensionsUnion, Measure, Metric
from dictum_core.model.dicts import DimensionDict, MeasureDict, MetricDict
from dictum_core.model.scalar import transforms as scalar_transforms
from dictum_core.model.table import RelatedTable, Table, TableFilter
from dictum_core.model.time import dimensions as time_dimensions

displayed_fields = {"id", "name", "description", "type", "format", "missing"}

table_calc_fields = displayed_fields | {"str_expr"}


class Model:
    def __init__(self, model: schema.Model):
        self.name = model.name
        self.description = model.description
        self.locale = model.locale

        self.tables: Dict[str, Table] = {}
        self.measures = MeasureDict()
        self.dimensions = DimensionDict()
        self.metrics = MetricDict()
        self.scalar_transforms = scalar_transforms

        self.theme = model.theme

        # add unions
        for union in model.unions.values():
            self.dimensions[union.id] = DimensionsUnion(
                **union.dict(include=displayed_fields)
            )

        # add all tables, their relationships and calculations
        for config_table in model.tables.values():
            table = self.create_table(config_table)
            self.tables[table.id] = table

            # add table dimensions
            for dimension in config_table.dimensions.values():
                self.add_dimension(dimension, table)

            # add table measures
            for measure in config_table.measures.values():
                self.add_measure(measure, table)

        # add detached dimensions
        for dimension in model.dimensions.values():
            table = self.tables[dimension.table]
            self.add_dimension(dimension, table)

        # add metrics
        for metric in model.metrics.values():
            self.add_metric(metric)

        # add measure backlinks
        for table in self.tables.values():
            for measure in table.measures.values():
                for target in table.allowed_join_paths:
                    target.measure_backlinks[measure.id] = table

        # add default time dimensions
        self.dimensions.update(time_dimensions)

    def create_table(self, table: schema.Table):
        result = Table(
            **table.dict(include={"id", "source", "description", "primary_key"})
        )
        for related in table.related.values():
            result.related[related.alias] = RelatedTable(
                parent=result,
                tables=self.tables,
                **related.dict(
                    include={"str_table", "str_related_key", "foreign_key", "alias"}
                ),
            )
        result.filters = [TableFilter(str_expr=f, table=result) for f in table.filters]
        return result

    def add_measure(self, measure: schema.Measure, table: Table) -> Measure:
        result = Measure(
            table=table,
            **measure.dict(include=table_calc_fields | {"str_filter", "str_time"}),
        )
        if measure.metric:
            self.metrics.add(Metric.from_measure(measure, self))
        table.measures.add(result)
        self.measures.add(result)

    def add_dimension(self, dimension: schema.Dimension, table: Table) -> Dimension:
        result = Dimension(
            table=table,
            **dimension.dict(include=table_calc_fields),
        )
        table.dimensions[result.id] = result
        if dimension.union is not None:
            if dimension.union in table.dimensions:
                raise KeyError(
                    f"Duplicate union dimension {dimension.union} "
                    f"on table {table.id}"
                )
            table.dimensions[dimension.union] = dataclasses.replace(
                result, id=dimension.union, is_union=True
            )
        self.dimensions.add(result)

    def add_metric(self, metric: schema.Metric):
        if metric.table is not None:
            # table is specified, treat as that table's measure
            table = self.tables.get(metric.table)
            measure = schema.Measure(**metric.dict(by_alias=True), metric=True)
            return self.add_measure(measure, table)

        # no, it's a real metric
        self.metrics[metric.id] = Metric(
            model=self,
            **metric.dict(include=table_calc_fields),
        )

    # def suggest_metrics(self, query: schema.Query) -> List[Measure]:
    #     """Suggest a list of possible metrics based on a query.
    #     Only metrics that can be used with all the dimensions from the query
    #     """
    #     result = []
    #     query_dims = set(r.dimension.id for r in query.dimensions)
    #     for metric in self.metrics.values():
    #         if metric.id in query.metrics:
    #             continue
    #         allowed_dims = set(d.id for d in metric.dimensions)
    #         if query_dims < allowed_dims:
    #             result.append(metric)
    #     return sorted(result, key=lambda x: (x.name))

    # def suggest_dimensions(self, query: schema.Query) -> List[Dimension]:
    #     """Suggest a list of possible dimensions based on a query. Only dimensions
    #     shared by all measures that are already in the query.
    #     """
    #     dims = set(self.dimensions) - set(r.dimension.id for r in query.dimensions)
    #     for request in query.metrics:
    #         metric = self.metrics.get(request.metric.id)
    #         for measure in metric.measures:
    #             dims = dims & set(measure.table.allowed_dimensions)
    #     return sorted([self.dimensions[d] for d in dims], key=lambda x: x.name)

    # def get_range_computation(self, dimension_id: str) -> Computation:
    #     """Get a computation that will compute a range of values for a given
    #       dimension.
    #     This is seriously out of line with what different parts of computation mean,
    #     so maybe we need to give them more abstract names.
    #     """
    #     dimension = self.dimensions.get(dimension_id)
    #     table = dimension.table
    #     min_, max_ = dimension.prepare_range_expr([table.id])
    #     return Computation(
    #         queries=[
    #             AggregateQuery(
    #                 join_tree=AggregateQuery(table=table, identity=table.id),
    #                 aggregate={"min": min_, "max": max_},
    #             )
    #         ]
    #     )

    # def get_values_computation(self, dimension_id: str) -> Computation:
    #     """Get a computation that will compute a list of unique possible values for
    #     this dimension.
    #     """
    #     dimension = self.dimensions.get(dimension_id)
    #     table = dimension.table
    #     return Computation(
    #         queries=[
    #             AggregateQuery(
    #                 join_tree=AggregateQuery(table=table, identity=table.id),
    #                 groupby={"values": dimension.prepare_expr([table.id])},
    #             )
    #         ]
    #     )

    def get_currencies_for_query(self, query: schema.Query):
        currencies = set()
        for request in query.metrics:
            metric = self.metrics.get(request.metric.id)
            if metric.format is not None and metric.format.currency is not None:
                currencies.add(metric.format.currency)
        for request in query.dimensions:
            dimension = self.dimensions.get(request.dimension.id)
            if dimension.format is not None and dimension.format.currency is not None:
                currencies.add(dimension.format.currency)
        return currencies



from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field, validator

from dictum_core.schema import utils
from dictum_core.schema.id import ID
from dictum_core.schema.model.calculations import Dimension, Measure


class RelatedTable(BaseModel):
    str_table: str = Field(alias="table")
    foreign_key: str
    alias: ID = Field(alias="id")
    str_related_key: Optional[str] = Field(alias="related_key")


class Table(BaseModel):
    id: ID
    description: Optional[str]
    source: Union[str, Dict]
    primary_key: Optional[str]
    related: Dict[ID, RelatedTable] = {}
    measures: Dict[ID, Measure] = {}
    dimensions: Dict[ID, Dimension] = {}
    filters: List[str] = []

    set_related_ids = validator("related", allow_reuse=True, pre=True)(utils.set_ids)
    set_measures_ids = validator("measures", allow_reuse=True, pre=True)(utils.set_ids)
    set_dimensions_ids = validator("dimensions", allow_reuse=True, pre=True)(
        utils.set_ids
    )



from typing import Optional

from pydantic import Field

from dictum_core.schema.id import ID
from dictum_core.schema.model.format import Formatted
from dictum_core.schema.model.type import Type


class Transform(Formatted):
    id: ID
    name: str
    description: Optional[str]
    args: list = []
    str_expr: str = Field(..., alias="expr")
    return_type: Optional[Type]



from dictum_core.model.calculations import Dimension, DimensionsUnion, Measure, Metric
from dictum_core.model.model import Model
from dictum_core.model.table import RelatedTable, Table

__all__ = [
    "Dimension",
    "DimensionsUnion",
    "Measure",
    "Metric",
    "Model",
    "RelatedTable",
    "Table",
]



import pytest
from lark import ParseError, Token, Tree, UnexpectedToken

from dictum_core.model.expr.introspection import get_expr_kind, get_expr_total_function
from dictum_core.model.expr.parser import parse_expr, parser


def test_integer_number():
    assert parse_expr("42").children[0] == Token("INTEGER", "42")
    assert parse_expr("3.14").children[0] == Token("FLOAT", "3.14")


def test_boolean():
    assert parse_expr("TrUe").children[0] == Token("TRUE", "TrUe")
    assert parse_expr("fAlSe").children[0] == Token("FALSE", "fAlSe")


def test_and_whitespace():
    """Test that whitespace around binary keyword operators (and/or) works the same as
    in Python.
    """
    tree = Tree("AND", [Token("INTEGER", "1"), Token("INTEGER", "1")])
    assert parse_expr("1 and 1").children[0] == tree
    assert parse_expr("(1)and(1)").children[0] == tree
    assert parse_expr("1and(1)").children[0] == tree
    with pytest.raises(UnexpectedToken):
        parse_expr("1and1")


def test_column():
    """Column syntax."""
    assert parse_expr("col").children[0] == Tree("column", ["col"])
    assert parser.parse("related.col").children[0] == Tree("column", ["related", "col"])
    assert parser.parse("related.other.col").children[0] == Tree(
        "column", ["related", "other", "col"]
    )


def test_measure():
    """Referencing other calculation"""
    assert parse_expr("$calc").children[0] == Tree("measure", ["calc"])
    with pytest.raises(ParseError):
        parse_expr("$table.calc")


def test_dimension():
    """Referencing other dimensions."""
    assert parse_expr(":dim").children[0] == Tree("dimension", ["dim"])
    with pytest.raises(ParseError):
        parse_expr(":table.dim")


def test_mul():
    """Test that multiplication is parsed.
    There was a bug where mul wasn't added to the grammar correctly.
    """
    assert parser.parse("2 * 2").children[0] == Tree(
        "mul", [Token("INTEGER", "2"), Token("INTEGER", "2")]
    )


def test_ne():
    """Test that both != and <> is supported."""
    tree = Tree("ne", [Token("INTEGER", "1"), Token("INTEGER", "1")])
    assert parser.parse("1 <> 1").children[0] == tree
    assert parser.parse("1 != 1").children[0] == tree


def test_fdiv():
    """Test that floor division operator (//) is replaced with a function call,
    e.g. 7 // 3 -> floor(7 / 3)
    """
    assert parse_expr("7 // 3").children[0] == Tree(
        "call", ["floor", Tree("div", [Token("INTEGER", "7"), Token("INTEGER", "3")])]
    )


def test_unquote():
    """Test that string values are unquoted."""
    assert parse_expr("'test'").children[0].value == "test"


def test_in():
    assert parse_expr("x in (1, 2, 3)").children[0] == Tree(
        "IN",
        [
            Tree("column", ["x"]),
            Token("INTEGER", "1"),
            Token("INTEGER", "2"),
            Token("INTEGER", "3"),
        ],
    )
    assert parse_expr("x in (0)").children[0] == Tree(
        "IN", [Tree("column", ["x"]), Token("INTEGER", "0")]
    )
    assert parse_expr("x > 0 and y in (0)").children[0] == Tree(
        "AND",
        [
            Tree("gt", [Tree("column", ["x"]), Token("INTEGER", "0")]),
            Tree("IN", [Tree("column", ["y"]), Token("INTEGER", "0")]),
        ],
    )
    assert parse_expr("y in (0) and x > 0").children[0] == Tree(
        "AND",
        [
            Tree("IN", [Tree("column", ["y"]), Token("INTEGER", "0")]),
            Tree("gt", [Tree("column", ["x"]), Token("INTEGER", "0")]),
        ],
    )
    assert parse_expr("x or 2 + 2 in (4)").children[0] == Tree(
        "OR",
        [
            Tree("column", ["x"]),
            Tree(
                "IN",
                [
                    Tree("add", [Token("INTEGER", "2"), Token("INTEGER", "2")]),
                    Token("INTEGER", "4"),
                ],
            ),
        ],
    )


def test_case_abbr():
    assert parse_expr("case when x > 0 then 1 end").children[0] == Tree(
        "case",
        [
            Tree("gt", [Tree("column", ["x"]), Token("INTEGER", "0")]),
            Token("INTEGER", "1"),
        ],
    )


def test_case_full():
    assert parse_expr("case when x > 0 then 1 else -1 end").children[0] == Tree(
        "case",
        [
            Tree(
                "gt",
                [
                    Tree("column", ["x"]),
                    Token("INTEGER", "0"),
                ],
            ),
            Token("INTEGER", "1"),
            Token("INTEGER", "-1"),
        ],
    )


def test_case_multi():
    assert parse_expr("case when x > 0 then 1 when x < 0 then -1 else 0 end").children[
        0
    ] == Tree(
        "case",
        [
            Tree("gt", [Tree("column", ["x"]), Token("INTEGER", "0")]),
            Token("INTEGER", "1"),
            Tree("lt", [Tree("column", ["x"]), Token("INTEGER", "0")]),
            Token("INTEGER", "-1"),
            Token("INTEGER", "0"),
        ],
    )


def test_arg():
    assert parse_expr("@ >= 10").children[0] == Tree(
        "ge", [Token("ARG", "@"), Token("INTEGER", "10")]
    )


def test_coalesce():
    assert parse_expr("test", missing=0).children[0] == Tree(
        "call", ["coalesce", Tree("column", ["test"]), Token("INTEGER", "0")]
    )


def test_null():
    assert parse_expr("x is null").children[0] == Tree(
        "isnull", [Tree("column", ["x"])]
    )
    assert parse_expr("x is not null").children[0] == Tree(
        "NOT", [Tree("isnull", [Tree("column", ["x"])])]
    )


def test_not_precedence():
    assert parse_expr("not if(x > 0, 1, 0) = 0").children[0] == Tree(
        "NOT",
        [
            Tree(
                "eq",
                [
                    Tree(
                        "call",
                        [
                            "IF",
                            Tree("gt", [Tree("column", ["x"]), Token("INTEGER", "0")]),
                            Token("INTEGER", "1"),
                            Token("INTEGER", "0"),
                        ],
                    ),
                    Token("INTEGER", "0"),
                ],
            )
        ],
    )


def test_expr_kind():
    assert get_expr_kind(parse_expr("1")) == "scalar"
    assert get_expr_kind(parse_expr("'a'")) == "scalar"

    assert get_expr_kind(parse_expr("col")) == "column"
    assert get_expr_kind(parse_expr("col + 1")) == "column"
    assert get_expr_kind(parse_expr("col * 2")) == "column"

    assert get_expr_kind(parse_expr("sum(col)")) == "aggregate"
    assert get_expr_kind(parse_expr("count(col) - 1")) == "aggregate"
    assert get_expr_kind(parse_expr("min(col) * 100")) == "aggregate"
    assert get_expr_kind(parse_expr("max(col) / 100")) == "aggregate"
    assert get_expr_kind(parse_expr("floor(sum(col - 1) + 1) / 10")) == "aggregate"
    assert get_expr_kind(parse_expr("case when max(col) > 1 then 1 else 0 end"))

    assert get_expr_kind(parse_expr("sum(col) in (1, 2, 3)")) == "aggregate"
    assert get_expr_kind(parse_expr("col in (1, 2, 3)")) == "column"
    assert get_expr_kind(parse_expr("1 in (1, 2, 3)")) == "scalar"


def test_expr_kind_invalid():
    with pytest.raises(ValueError, match="error in expression"):
        get_expr_kind(parse_expr("sum(col) + col"))
    with pytest.raises(ValueError, match="error in expression"):
        get_expr_kind(parse_expr("case when col = 1 then sum(col) end"))
    with pytest.raises(ValueError, match="error in expression"):
        get_expr_kind(parse_expr("case when col = 1 then sum(col) end"))
    with pytest.raises(ValueError, match="error in expression"):
        get_expr_kind(parse_expr("col in (sum(col), 1, 2)"))


def test_expr_total_function():
    assert get_expr_total_function(parse_expr("sum(col)")) == "sum"
    assert get_expr_total_function(parse_expr("sum(col + 1)")) == "sum"
    assert get_expr_total_function(parse_expr("count(col)")) == "sum"
    assert get_expr_total_function(parse_expr("max(col)")) == "max"
    assert get_expr_total_function(parse_expr("min(col)")) == "min"
    assert get_expr_total_function(parse_expr("avg(col)")) is None
    assert get_expr_total_function(parse_expr("countd(col)")) is None
    assert get_expr_total_function(parse_expr("sum(col) + 1")) is None



import dataclasses
from datetime import date, datetime

from lark import Token, Tree


def repr_expr_constant(val):
    if isinstance(val, str):
        return f"'{val}'"
    return str(val)


def value_to_token(value):
    if value is True:
        return Token("TRUE", "True")
    if value is False:
        return Token("FALSE", "False")
    if isinstance(value, int):
        return Token("INTEGER", str(value))
    if isinstance(value, float):
        return Token("NUMBER", str(value))
    if isinstance(value, str):
        return Token("STRING", value)
    if isinstance(value, date):
        return Token("DATE", value.strftime(r"%Y-%m-%d"))
    if isinstance(value, datetime):
        return Token("DATETIME", value.strftime(r"%Y-%m-%d %H:%M:%S"))
    raise ValueError("Token values must by integers, floats or strings")


def column_expr(name: str) -> Tree:
    return Tree("expr", [Tree("column", [None, name])])


def subselect_column(column):
    return dataclasses.replace(
        column, expr=Tree("expr", [Tree("column", [None, column.name])])
    )



from dataclasses import dataclass
from typing import Dict, List, Literal, Optional

from dictum_core import schema


@dataclass
class ExecutedQuery:
    query: str
    time: float


AltairTimeUnit = Literal[
    "year",
    "yearquarter",
    "yearmonth",
    "yearmonthdate",
    "yearmonthdatehours",
    "yearmonthdatehoursminutes",
    "yearmonthdatehoursminutesseconds",
]

DisplayColumnKind = Literal["metric", "dimension"]


@dataclass
class DisplayInfo:
    """Information for the displaying code:
    either data formatter or Altair
    """

    display_name: str
    column_name: str
    format: schema.FormatConfig
    kind: DisplayColumnKind
    type: Optional[schema.Type] = None
    keep_display_name: bool = False
    altair_time_unit: Optional[str] = None


@dataclass
class Result:
    data: List[dict]
    display_info: Dict[str, DisplayInfo]
    executed_queries: List[ExecutedQuery]



from abc import ABC, abstractmethod
from enum import Enum

from lark import Transformer, Tree
from lark.exceptions import VisitError


class AbstractExprTransformer(Transformer, ABC):
    @abstractmethod
    def FLOAT(self, value: str):
        ...

    @abstractmethod
    def INTEGER(self, value: str):
        ...

    @abstractmethod
    def STRING(self, value: str):
        ...

    @abstractmethod
    def TRUE(self, _):
        ...

    @abstractmethod
    def FALSE(self, _):
        ...

    @abstractmethod
    def column(self, children: list):
        ...

    @abstractmethod
    def measure(self, children: list):
        ...

    @abstractmethod
    def dimension(self, children: list):
        ...

    @abstractmethod
    def call(self, children: list):
        ...

    @abstractmethod
    def exp(self, children: list):
        ...

    @abstractmethod
    def neg(self, children: list):
        ...

    @abstractmethod
    def fdiv(self, children: list):
        ...

    @abstractmethod
    def div(self, children: list):
        ...

    @abstractmethod
    def mul(self, children: list):
        ...

    @abstractmethod
    def mod(self, children: list):
        ...

    @abstractmethod
    def add(self, children: list):
        ...

    @abstractmethod
    def sub(self, children: list):
        ...

    @abstractmethod
    def gt(self, children: list):
        ...

    @abstractmethod
    def ge(self, children: list):
        ...

    @abstractmethod
    def lt(self, children: list):
        ...

    @abstractmethod
    def le(self, children: list):
        ...

    @abstractmethod
    def eq(self, children: list):
        ...

    @abstractmethod
    def ne(self, children: list):
        ...

    @abstractmethod
    def isnull(self, children: list):
        ...

    @abstractmethod
    def IN(self, children: list):
        ...

    @abstractmethod
    def NOT(self, children: list):
        ...

    @abstractmethod
    def AND(self, children: list):
        ...

    @abstractmethod
    def OR(self, children: list):
        ...

    @abstractmethod
    def case(self, children: list):
        ...

    @abstractmethod
    def expr(self, children: list):
        ...


class ExprKind(str, Enum):
    scalar = "scalar"
    aggregate = "aggregate"
    column = "column"


def scalar(self, _):
    return ExprKind.scalar


def _infer_kind(items: list):
    items = set(items)

    if len(items) == 1:
        return items.pop()

    if items == {ExprKind.column, ExprKind.scalar}:
        return ExprKind.column

    if items == {ExprKind.aggregate, ExprKind.scalar}:
        return ExprKind.aggregate

    if items & {ExprKind.aggregate, ExprKind.column}:
        raise ValueError("Mixing aggregates and non-aggregates in expression")

    raise ValueError(f"Unknown combination of kinds: {items}")


def infer_kind(self, children: list):
    return _infer_kind(children)


aggregate_functions = {
    "sum",
    "count",
    "countd",
    "min",
    "max",
    "avg",
}


class ExprKindTransformer(AbstractExprTransformer):
    """Checks that the expression is valid, namely that it's not mixing
    aggregate and non-aggregate parts, e.g.
        column + sum(another_column) isn't valid

    parts of expressions can be
        - scalar (constants)
        - aggregate (result of aggregate function)
        - column (a column reference or a scalar operation on it)

    things inside aggregate function calls must be scalar or column
    operations with aggregates must be scalar or aggregate

    scalar + scalar gives scalar
    scalar + column gives column

    agg(scalar) gives aggregate
    agg(column) gives aggregate
    agg(aggregate) is invalid

    aggregate + scalar gives aggregate
    aggregate + column is invalid
    """

    FLOAT = scalar
    INTEGER = scalar
    STRING = scalar
    TRUE = scalar
    FALSE = scalar

    def column(self, _):
        return ExprKind.column

    def measure(self, _):
        return ExprKind.aggregate

    def dimension(self, _):
        return ExprKind.column

    def call(self, children: list):
        fn, *args = children
        if fn in aggregate_functions:
            if args and args[0] == ExprKind.aggregate:
                raise ValueError(
                    f"Aggregate function {fn} expects a scalar "
                    "or a column expression argument"
                )
            return ExprKind.aggregate
        return _infer_kind(args)

    exp = infer_kind
    neg = infer_kind
    fdiv = infer_kind
    div = infer_kind
    mul = infer_kind
    mod = infer_kind
    add = infer_kind
    sub = infer_kind
    gt = infer_kind
    ge = infer_kind
    lt = infer_kind
    le = infer_kind
    eq = infer_kind
    ne = infer_kind
    eq = infer_kind
    ne = infer_kind
    isnull = infer_kind
    IN = infer_kind
    NOT = infer_kind
    AND = infer_kind
    OR = infer_kind

    def case(self, children: list):
        return _infer_kind(children)

    def expr(self, children: list):
        return children[0]


expr_kind_transformer = ExprKindTransformer()


def get_expr_kind(expr: Tree) -> ExprKind:
    try:
        return expr_kind_transformer.transform(expr)
    except VisitError as e:
        raise ValueError(f"error in expression {e.obj.data}: {e.orig_exc}")


class TotalFunctionExprTransformer(Transformer):
    """For aggregate expressions we need to figure out which function (if any) to use
    for calculating its totals.

    We know for sure which function to use if an aggregate is just a bare function:
        sum -> sum
        count -> sum
        max -> max
        min -> min
    For all other cases, e.g. countd, sum(..) + 1 or sum(..) / sum(..), return None
    (no function), so that the system can calculate the total some other way (subquery).

    Mainly to be used in total and percent transforms.
    """

    def expr(self, children: list):
        return children[0]

    def call(self, children: list):
        fn, *_ = children
        if fn == "count":
            return "sum"
        if fn in {"sum", "min", "max"}:
            return fn
        # TODO: maybe there are more cases where the function is known?
        return None  # unknown

    def __default__(self, *_):
        return None


total_function_expr_transformer = TotalFunctionExprTransformer()


def get_expr_total_function(expr: Tree) -> str:
    return total_function_expr_transformer.transform(expr)



from typing import List, Optional

from lark import Token, Tree

import dictum_core.engine.engine
from dictum_core.engine.aggregate_query_builder import AggregateQueryBuilder
from dictum_core.engine.computation import Column
from dictum_core.engine.operators import (
    FilterOperator,
    MaterializeOperator,
    MergeOperator,
    QueryOperator,
    RecordsFilterOperator,
)
from dictum_core.engine.result import DisplayInfo
from dictum_core.model import Metric, Model
from dictum_core.schema import (
    FormatConfig,
    Query,
    QueryDimension,
    QueryDimensionRequest,
    QueryMetric,
    QueryMetricRequest,
    QueryTableTransform,
)


class AddMetric:
    """Add a metric to a merge.
    - For each measure of the metric
      - Add the measure aggregation to the merge (if it's not already there)
    - Add the necessary calculation to the merge.
    """

    def __init__(
        self,
        request: QueryMetricRequest,
        builder: Optional[AggregateQueryBuilder] = None,
    ):
        self.request = request
        self.builder = builder

    @property
    def model(self) -> Model:
        return self.builder.model

    @property
    def metric(self) -> Metric:
        return self.model.metrics.get(self.request.metric.id)

    def add_measures(self, merge: MergeOperator) -> MergeOperator:
        existing_measures = {
            q.input._aggregate[0].name
            for q in merge.inputs
            if isinstance(q, QueryOperator)
        }
        for measure in self.metric.measures:
            if measure.id not in existing_measures:
                aggregation = self.builder.get_aggregation(measure)
                merge.add_query(QueryOperator(aggregation))
        return merge

    def get_terminal_for_query(self, query: Query) -> MergeOperator:
        engine = dictum_core.engine.engine.Engine(self.model)
        return engine.get_terminal(query=query)

    def __call__(self, merge: MergeOperator):
        merge = self.add_measures(merge)
        column = Column(
            name=self.request.digest,
            expr=self.metric.merged_expr,
            type=self.metric.type,
            display_info=DisplayInfo(
                display_name=(
                    self.metric.name if not self.request.alias else self.request.alias
                ),
                column_name=self.request.name,
                format=self.metric.format,
                type=self.metric.type,
                kind="metric",
            ),
        )
        merge.metrics.append(column)
        return merge


transforms = {}


class AddTransformedMetric(AddMetric):
    def __init_subclass__(cls) -> None:
        if hasattr(cls, "id"):
            transforms[cls.id] = cls

    @property
    def transform(self):
        return self.request.metric.transforms[0]

    def get_transform_terminal(
        self,
        request: Optional[QueryMetricRequest] = None,
        dimensions: Optional[List[QueryDimension]] = None,
        filters: Optional[List[QueryDimension]] = None,
    ) -> MergeOperator:
        # create our own query based on the transform
        # dimensions are of + within if not specified
        if request is None:
            request = self.request

        if dimensions is None:
            dimensions = [*self.transform.of, *self.transform.within]

        dimensions = [QueryDimensionRequest(dimension=d) for d in dimensions]

        if filters is None:
            filters = self.builder.filters

        # ignore alias and transforms
        query = Query(
            metrics=[QueryMetricRequest(metric=QueryMetric(id=request.metric.id))],
            dimensions=dimensions,
            filters=filters,
        )
        result = self.get_terminal_for_query(query)

        # replace with the request's expected column name
        result.metrics[0].name = request.digest

        return result

    def get_unlisted_query_dimensions(self) -> List[QueryDimension]:
        """Get a list of requests for dimensions that appear neither
        in "of" nor in "within".
        """
        digests = {d.digest for d in self.transform.of} | {
            d.digest for d in self.transform.within
        }
        return [
            r.dimension
            for r in self.builder.dimensions
            if r.dimension.digest not in digests
        ]


limit_transforms = {}


class AddLimit(AddTransformedMetric):
    """A metric adder user for transforms that are only used in query's limit clause"""

    def __init_subclass__(cls) -> None:
        if hasattr(cls, "id"):
            limit_transforms[cls.id] = cls


class AddTopBottomLimit(AddLimit):
    """Adds a top and or bottom limit to the computation.

    The general logic is that we need to construct a query
    with the requested metric, wrap the metric in row_number
    and only leave the rows where row_number is <= (or >=)
    than the specified limit. Then the dimension tuples of
    the result are taken and put into the WHERE clause of
    each measure query. So basically we first compute which
    dimension values must be in the result and then filter
    by these values.

    Multiple such queries are computed in parallel saving
    time, but generally this requires at least two separate
    backend queries to be run in sequence.

    TODO: optimize by computing everything in the database
    when there's only one limit (with an inner join to subquery)
    """

    ascending: bool

    def __init__(
        self, metric: QueryMetric, builder: Optional[AggregateQueryBuilder] = None
    ):
        super().__init__(QueryMetricRequest(metric=metric), builder)

    def wrap_in_row_number(self, merge: MergeOperator, within: List[QueryDimension]):
        """Add row_number to the terminal operator."""
        column = merge.columns[-1]
        metric_expr = column.expr.children[0]
        order_by = Tree(
            "order_by",
            [Tree("order_by_item", [metric_expr, self.ascending])],
        )

        within_names = {r.digest for r in within}
        partition_by = Tree(
            "partition_by", [c.expr for c in merge.columns if c.name in within_names]
        )

        column.expr = Tree(
            "expr",
            [
                Tree(
                    "call_window",
                    ["row_number", metric_expr, partition_by, order_by, None],
                )
            ],
        )
        column.name = "__row_number"
        column.type = "int"
        return merge

    def __call__(self, merge: MergeOperator):
        # if "of" is empty then "of" is everything that's not "within"
        if len(self.transform.of) == 0:
            # TODO: do not mutate the original request
            self.transform.of = self.get_unlisted_query_dimensions()

        terminal = self.get_transform_terminal()
        terminal = self.wrap_in_row_number(terminal, self.transform.within)

        # add <= filter by row_number
        terminal = FilterOperator(
            input=terminal,
            conditions=[
                Tree(
                    "le",
                    [
                        Tree("column", [None, "__row_number"]),
                        Token("INTEGER", self.transform.args[0]),
                    ],
                )
            ],
        )

        # if the inputs to the merge are TuplesFilters, it means that
        # this top isn't the first, so we just add the terminal to
        # materialize
        if all(isinstance(i, RecordsFilterOperator) for i in merge.inputs):
            materialized = merge.inputs[0].materialized
            materialized.inputs.append(terminal)
            return merge

        # otherwize create MaterializeOperator and add a TuplesFilter
        # before each query

        # terminal Merge -> Materialize
        materialized = MaterializeOperator([terminal])

        # add TuplesFilter to each input of the original merge
        # (excluding merges which mean transformed metrics)
        new_inputs = []
        for input in merge.inputs:
            new_inputs.append(
                RecordsFilterOperator(
                    query=input, materialized=materialized, drop_last_column=True
                )
            )
        merge.inputs = new_inputs

        return merge


class AddTopLimit(AddTopBottomLimit):
    id = "top"
    name = "Top"
    ascending: bool = False


class AddBottomLimit(AddTopBottomLimit):
    id = "bottom"
    name = "Bottom"
    ascending: bool = True


class AddTotalMetric(AddTransformedMetric):
    """Totals are computed as a merge added as a source to the merge."""

    id = "total"
    name = "Total"

    def get_column(self) -> Column:
        return

    def add_base_dimensions(self, merge: MergeOperator):
        """To have all the dimensions in the merge for sure, we have to add the measures
        for the original metric in case the total is the only one requested
        """
        merge = AddMetric(request=self.request, builder=self.builder)(merge)
        merge.metrics.pop(-1)  # remove the last metric, we only need the query

    def get_total_metric_request(
        self, dimensions: Optional[List[QueryDimension]] = None
    ) -> QueryMetricRequest:
        if dimensions is None:
            dimensions = [*self.transform.of, *self.transform.within]
        return QueryMetricRequest(
            metric=QueryMetric(
                id=self.metric.id,
                transforms=[QueryTableTransform(id="total", within=dimensions)],
            )
        )

    def __call__(self, merge: MergeOperator):
        # TODO: optimize by using get_expr_total_function
        # if the total function is defined (expr is additive),
        # then use a window function instead of adding another
        # query to the merge

        self.add_base_dimensions(merge)
        terminal = self.get_transform_terminal(request=self.get_total_metric_request())
        merge.add_merge(terminal)
        column = Column(
            name=self.request.name,
            expr=Tree("expr", [Tree("column", [None, terminal.metrics[0].name])]),
            type=self.metric.type,
            display_info=DisplayInfo(
                display_name=(
                    f"{self.metric.name} (Total)"
                    if self.request.alias is None
                    else self.request.alias
                ),
                column_name=self.request.name,
                format=self.metric.format,
                type=self.metric.type,
                kind="metric",
            ),
        )
        merge.metrics.append(column)
        return merge


class AddPercentMetric(AddTotalMetric):
    id = "percent"
    name = "Percent"

    def get_metric_expr(self) -> Tree:
        return self.metric.merged_expr.children[0]

    def __call__(self, merge: MergeOperator):
        """Percent actually includes two separate queries:
        - the value for the numerator
        - the value for the denominator

        cases (all by country, city, region):

        1. bare percent — all rows add up to 100%
            divide metric by total
        2. percent within (country) — all tuples within country add up to 100%
            so "of" is evetyhing that's not "within"
            just divide metric by total like above
        3. percent of (city) within (country) — ignores region
            numerator: total within (city, country)
            denominator: total within (country)
        4. percent of (city) — all cities within each tuple add up to 100%
            "within" is everything that's not "of", i.e.
            percent of (city) within (country, region)
            This case gets reduced to the previous case:
                numerator: total within (city, country, region)
                denominator: total within (country, region)
            The numerator is already computed — it's the metric itself, so we can just
            divide the metric by total within (country, region)
        """
        self.add_base_dimensions(merge)

        # figure out which case we're dealing with
        # based on that info, we can construct queries for the numerator and denominator
        if len(self.transform.of) == 0:
            # cases 1 and 2, just divide metric by total
            numerator = self.get_metric_expr()
            dimensions = [*self.transform.of, *self.transform.within]
            denominator_terminal = self.get_transform_terminal(
                request=self.get_total_metric_request(dimensions),
                dimensions=dimensions,
            )
            denominator = Tree("column", [None, denominator_terminal.metrics[0].name])
            merge.add_merge(denominator_terminal)
        elif len(self.transform.within) == 0:
            # case 4, of present, within absent
            dimensions = self.get_unlisted_query_dimensions()
            denominator_terminal = self.get_transform_terminal(
                request=self.get_total_metric_request(dimensions),
                dimensions=dimensions,
            )
            denominator = Tree("column", [None, denominator_terminal.metrics[0].name])
            numerator = self.get_metric_expr()
            merge.add_merge(denominator_terminal)
        else:
            # case 3, we need to actually construct two terminals
            # like in total, add the metric query
            merge = AddMetric(request=self.request, builder=self.builder)(merge)
            merge.metrics.pop(-1)  # remove the last metric, we only need the query

            # add numerator
            dimensions = [*self.transform.of, *self.transform.within]
            numerator_terminal = self.get_transform_terminal(
                request=self.get_total_metric_request(dimensions),
                dimensions=self.transform.of + self.transform.within,
            )
            numerator = Tree("column", [None, numerator_terminal.metrics[0].name])
            merge.add_merge(numerator_terminal)

            denominator_terminal = self.get_transform_terminal(
                request=self.get_total_metric_request(self.transform.within),
                dimensions=self.transform.within,
            )
            denominator = Tree("column", [None, denominator_terminal.metrics[0].name])
            merge.add_merge(denominator_terminal)

        column = Column(
            name=self.request.name,
            expr=Tree(
                "expr",
                [Tree("div", [numerator, denominator])],
            ),
            type="float",
            display_info=DisplayInfo(
                display_name=(
                    f"{self.metric.name} (%)"
                    if self.request.alias is None
                    else self.request.alias
                ),
                column_name=self.request.name,
                format=FormatConfig(kind="percent"),
                type=self.metric.type,
                kind="metric",
            ),
        )
        merge.metrics.append(column)
        return merge


class AddAdditivelyTransformedMetric(AddTransformedMetric):
    """Sum, max, min"""

    def get_column(self) -> Column:
        raise NotImplementedError

    def __call__(self, merge: MergeOperator):
        # we have to add the measures for the original metric in case this transform
        # is the only one requested
        merge = AddMetric(request=self.request, builder=self.builder)(merge)
        merge.metrics.pop(-1)  # remove the last metric, we only need the query

        merge.metrics.append(self.get_column())
        return merge


class AddSumMetric(AddAdditivelyTransformedMetric):
    id = "sum"
    name = "Sum"

    def get_column(self) -> Column:
        expr = self.metric.merged_expr.children[0]
        dimensions = self.transform.of + self.transform.within
        partition_by = Tree(
            "partition_by", [Tree("column", [None, d.digest]) for d in dimensions]
        )
        return Column(
            name=self.request.name,
            expr=Tree(
                "expr", [Tree("call_window", ["sum", expr, partition_by, None, None])]
            ),
            type=self.metric.type,
            display_info=DisplayInfo(
                display_name=(
                    self.metric.name
                    if self.request.alias is None
                    else self.request.alias
                ),
                column_name=self.request.name,
                format=self.metric.format,
                type=self.metric.type,
                kind="metric",
            ),
        )



from IPython.core.magic import Magics, line_cell_magic, magics_class
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring


@magics_class
class QlMagics(Magics):
    def __init__(self, project, shell=None, **kwargs):
        super().__init__(shell=shell, **kwargs)
        self.project = project

    @line_cell_magic("ql")
    @magic_arguments()
    @argument(
        "-r",
        "--result",
        default=None,
        action="store",
        help="Store the resulting dataframe in a local variable",
    )
    @argument(
        "-f",
        "--format-result",
        action="store_true",
        default=False,
        help="Format the result returned to a local variable",
    )
    @argument(
        "-n",
        "--no-formatting",
        action="store_true",
        default=False,
        help="Display the result without formatting",
    )
    @argument(
        "query",
        default=None,
        action="store",
        nargs="*",  # the rest is query
    )
    def ql(self, line, cell=None):
        args = parse_argstring(self.ql, line)
        format = not args.no_formatting

        query = cell
        if cell is None:
            query = " ".join(args.query)

        if args.result is not None:
            result = self.project.ql(query).df(format=args.format_result)
            self.shell.user_ns.update({args.result: result})
            print(f"Returned {result.shape[0]} rows to {args.result}")
            return  # don't show the result in Jupiter

        result = self.project.ql(query).df(format=format)
        return result



import pytest

from dictum_core.model import Model
from dictum_core.model.table import RelatedTable, Table


def test_table_find_all_paths_chinook(chinook: Model):
    paths = list(chinook.tables["invoice_items"].find_all_paths())
    assert len(paths) == 8


def test_table_allowed_join_paths(chinook: Model):
    table = chinook.tables["invoice_items"]
    assert len(table.allowed_join_paths) == 8


def test_table_allowed_dimensions(chinook: Model):
    table = chinook.tables["invoice_items"]
    assert "artist" in table.allowed_dimensions
    assert "customer_country" in table.allowed_dimensions


def test_table_allowed_dimensions_union(chinook: Model):
    assert "country" not in chinook.tables.get("invoice_items").allowed_dimensions
    assert "country" in chinook.tables.get("customers").allowed_dimensions


@pytest.fixture(scope="module")
def t1() -> Table:
    """Table graph representing (hopefully) all possible table relatioship cases
    - Self join
    - Related self join
    - Directly related table
    - Relationship cycle
    - Duplicate join path

    ┌┐          ┌┐    ┌─────┐
    ▼┘          ▼┘    ▼     │
    t1 ─► t2 ─► t3 ─► t4 ─► t5
    │                       ▲
    └► t6 ──────────────────┘

    Returns: t1
    """

    t1 = Table(id="t1", source="test")
    t2 = Table(id="t2", source="test")
    t3 = Table(id="t3", source="test")
    t4 = Table(id="t4", source="test")
    t5 = Table(id="t5", source="test")
    t6 = Table(id="t6", source="test")
    tables = {"t1": t1, "t2": t2, "t3": t3, "t4": t4, "t5": t5, "t6": t6}

    def create_rel(parent, table):
        return RelatedTable(
            str_table=table.id,
            foreign_key=f"{table.id}_id",
            str_related_key="id",
            alias=table.id,
            parent=parent,
            tables=tables,
        )

    t1.related["t1"] = create_rel(t1, t1)
    t1.related["t2"] = create_rel(t1, t2)
    t1.related["t6"] = create_rel(t1, t6)
    t2.related["t3"] = create_rel(t2, t3)
    t3.related["t3"] = create_rel(t3, t3)
    t3.related["t4"] = create_rel(t3, t4)
    t4.related["t5"] = create_rel(t4, t5)
    t5.related["t4"] = create_rel(t5, t4)
    t6.related["t5"] = create_rel(t6, t5)
    return t1


def test_find_all_paths(t1: Table):
    assert [p for _, p in t1.find_all_paths()] == [
        ["t2"],
        ["t2", "t3"],
        ["t2", "t3", "t4"],
        ["t2", "t3", "t4", "t5"],
        ["t6"],
        ["t6", "t5"],
        ["t6", "t5", "t4"],
    ]


def test_allowed_join_paths(t1: Table):
    paths = {t.id: path for t, path in t1.allowed_join_paths.items()}
    assert paths == {"t2": ["t2"], "t3": ["t2", "t3"], "t6": ["t6"], "t1": ["t1"]}



import pytest
from lark import Token, Tree

from dictum_core.engine import Column, DisplayInfo
from dictum_core.model.scalar import LiteralTransform, ScalarTransform, transforms
from dictum_core.schema import FormatConfig


@pytest.fixture(scope="function")
def col():
    return Column(
        name="test",
        type="int",
        expr=Tree("expr", [Token("INTEGER", "1")]),
    )


def test_transform():
    transform = ScalarTransform()
    assert transform.get_return_type("test") == "test"
    assert transform.get_format("test") == "test"
    with pytest.raises(NotImplementedError):
        transform.transform_expr(None)


def test_literal_transform(col: Column):
    class LiteralTest(LiteralTransform):
        expr = "@ > value"
        args = ["value"]
        return_type = "bool"

    transform = LiteralTest(0)
    assert transform._expr.children[0] == Tree(
        "gt", [Token("ARG", "@"), Tree("column", ["value"])]
    )
    assert transform.transform_expr(Token("TEST", None)) == Tree(
        "gt", [Token("TEST", None), Token("INTEGER", "0")]
    )
    assert transform(col) == Column(
        name="test",
        type="bool",
        expr=Tree(
            "expr",
            [
                Tree(
                    "gt",
                    [
                        Token("INTEGER", "1"),
                        Token("INTEGER", "0"),
                    ],
                )
            ],
        ),
    )


def test_booleans(col: Column):
    for key in ["eq", "ne", "gt", "ge", "lt", "le"]:
        transform = transforms[key](0)
        result = transform(col)
        assert result.type == "bool"
        assert result.expr.children[0].data == key
        assert result.expr.children[0].children == [
            Token("INTEGER", "1"),
            Token("INTEGER", "0"),
        ]


def test_nulls(col: Column):
    result = transforms["isnull"]()(col)
    assert result.type == "bool"
    assert result.expr == Tree("expr", [Tree("isnull", [Token("INTEGER", "1")])])

    result = transforms["isnotnull"]()(col)
    assert result.type == "bool"
    assert result.expr == Tree(
        "expr", [Tree("NOT", [Tree("isnull", [Token("INTEGER", "1")])])]
    )


def test_inrange(col: Column):
    result = transforms["inrange"](-1, 1)(col)
    assert result.type == "bool"
    assert result.expr == Tree(
        "expr",
        [
            Tree(
                "AND",
                [
                    Tree("ge", [Token("INTEGER", "1"), Token("INTEGER", -1)]),
                    Tree("le", [Token("INTEGER", "1"), Token("INTEGER", "1")]),
                ],
            )
        ],
    )


def test_in(col: Column):
    result = transforms["isin"](0, 1, 2)(col)
    assert result.type == "bool"
    assert result.expr.children[0] == Tree(
        "IN",
        [
            col.expr.children[0],
            Token("INTEGER", "0"),
            Token("INTEGER", "1"),
            Token("INTEGER", "2"),
        ],
    )


def test_datepart(col: Column):
    result = transforms["datepart"]("month")(col)
    assert result.type == "int"
    assert result.expr.children[0] == Tree(
        "call", ["datepart", "month", col.expr.children[0]]
    )

    result = transforms["month"]()(col)
    assert result.type == "int"
    assert result.expr.children[0] == Tree(
        "call", ["datepart", "month", col.expr.children[0]]
    )


def test_datetrunc(col: Column):
    col.display_info = DisplayInfo(
        display_name="xxx",
        column_name="xxx",
        format=FormatConfig(kind="date"),
        kind="dimension",
        altair_time_unit="month",
    )
    datetrunc = transforms["datetrunc"]("month")
    result = datetrunc(col)
    assert result.type == "datetime"
    assert result.expr.children[0] == Tree(
        "call", ["datetrunc", "month", col.expr.children[0]]
    )

    assert datetrunc.get_format("datetime").pattern is None
    assert datetrunc.get_format("datetime").skeleton == "yMMM"



from copy import deepcopy
from typing import List

from lark import Tree


def merged_expr(name: str):
    return Tree("expr", [Tree("column", [None, name])])


def prefixed_expr(expr: Tree, prefix: List[str]) -> Tree:
    """Prefix the expression with the given join path."""
    result = deepcopy(expr)
    for ref in result.find_data("column"):
        # skip first child: host table's name
        _, *path, field = ref.children
        ref.children = [*prefix, *path, field]
    return result


def prepare_expr(expr: Tree, prefix: List[str]) -> Tree:
    """Prepare the expression for query: turn prefixed path into .-delimited string"""
    expr = prefixed_expr(expr, prefix)
    for ref in expr.find_data("column"):
        *path, field = ref.children
        ref.children = [".".join(path), field]
    return expr



from dictum_core.engine.computation import (
    Column,
    Join,
    LiteralOrderItem,
    OrderItem,
    RelationalQuery,
)
from dictum_core.engine.engine import Engine
from dictum_core.engine.operators import (
    FilterOperator,
    InnerJoinOperator,
    MaterializeOperator,
    MergeOperator,
    Operator,
    QueryOperator,
    RecordsFilterOperator,
)
from dictum_core.engine.result import DisplayInfo, Result

__all__ = [
    "Column",
    "DisplayInfo",
    "Engine",
    "FilterOperator",
    "InnerJoinOperator",
    "Join",
    "LiteralOrderItem",
    "MaterializeOperator",
    "MergeOperator",
    "Operator",
    "OrderItem",
    "QueryOperator",
    "RelationalQuery",
    "Result",
    "RecordsFilterOperator",
]



from abc import ABC, abstractmethod


class DatediffCompilerMixin(ABC):
    """Mixin to calculate datediff from datepart and datetrunc implementations for
    systems that don't support datediff directly.

    The child class only needs to implement datediff_day
    """

    def datediff_year(self, start, end):
        return self.datepart("year", end) - self.datepart("year", start)

    def datediff_quarter(self, start, end):
        return self.datediff_year(start, end) * 4 + (
            self.datepart("quarter", end) - self.datepart("quarter", start)
        )

    def datediff_month(self, start, end):
        return self.datediff_year(start, end) * 12 + (
            self.datepart("month", end) - self.datepart("month", start)
        )

    def datediff_week(self, start, end):
        start_week = self.datetrunc("week", start)
        end_week = self.datetrunc("week", end)
        return self.tointeger(self.datediff_day(start_week, end_week) / 7)

    @abstractmethod
    def datediff_day(self, start, end):
        """This can't be implemented through other functions"""

    def datediff_hour(self, start, end):
        days = self.datediff_day(start, end)
        return days * 24 + (self.datepart("hour", end) - self.datepart("hour", start))

    def datediff_minute(self, start, end):
        hours = self.datediff_hour(start, end)
        return hours * 60 + (
            self.datepart("minute", end) - self.datepart("minute", start)
        )

    def datediff_second(self, start, end):
        minutes = self.datediff_minute(start, end)
        return minutes * 60 + (
            self.datepart("second", end) - self.datepart("second", start)
        )

    def datediff(self, part, start, end):
        fn = {
            "year": self.datediff_year,
            "quarter": self.datediff_quarter,
            "month": self.datediff_month,
            "week": self.datediff_week,
            "day": self.datediff_day,
            "hour": self.datediff_hour,
            "minute": self.datediff_minute,
            "second": self.datediff_second,
        }.get(part.lower())
        if fn is None:
            raise ValueError(
                "Valid values for datediff part are year, quarter, "
                "month, week, day, hour, minute, second — "
                f"got '{part}'."
            )
        return fn(start, end)



import os
from getpass import getpass
from itertools import chain
from pathlib import Path
from typing import Dict, Optional, Union

import yaml
from jinja2 import Template
from pydantic import BaseModel

from dictum_core.schema.id import ID
from dictum_core.schema.model import Model

root_items = {"tables", "metrics", "unions", "profiles"}


def _merge(target: dict, source: dict):
    result = target.copy()
    for k, v in source.items():
        if k in root_items:
            result[k] = _merge(result.get(k, {}), source[k])
            continue
        if k in result:
            raise KeyError(f"Duplicate key in project config: {k}")
        result[k] = v
    return result


class Env:
    def __init__(self):
        self.data = os.environ.copy()

    def __getattr__(self, key: str):
        try:
            return self.data[key]
        except KeyError:
            val = getpass(key)
            self.data[key] = val
            return val


env = Env()


def _load_yaml_template(path: Path):
    return yaml.load(Template(path.read_text()).render(env=env), Loader=yaml.SafeLoader)


def _load_dict_from_path(path: Union[str, Path]) -> dict:
    """Load a single YAML file or multiple YAML files in nested folders."""
    path = Path(path)
    if path.is_file():
        result = _load_yaml_template(path)
        if result is None:
            result = {}
        if not isinstance(result, dict):
            raise TypeError(f"Error loading {path}: top-level object must be a dict")
        return result

    if path.is_dir():
        result = {}
        for file in chain(path.glob("**/*.yml"), path.glob("**/*.yaml")):
            key = file.stem
            data = _load_yaml_template(file)
            if not isinstance(data, dict):
                raise TypeError(
                    f"Error loading {path}: top-level object must be a dict"
                )
            result[key] = data
        return result

    raise FileNotFoundError(f"Path {path} does not exist")


class Profile(BaseModel):
    type: str
    parameters: dict = {}


class Profiles(BaseModel):
    default_profile: str
    profiles: Dict[ID, Profile]


class Project(BaseModel):
    name: str
    description: Optional[str]
    locale: str = "en_US"
    currency: str = "USD"

    tables_path: str = "tables"
    metrics_path: str = "metrics"
    unions_path: str = "unions.yml"
    profiles_path: str = "profiles.yml"

    root: Path

    @classmethod
    def load(cls, path: Union[str, Path]):
        path = Path(path)
        if not path.is_dir():
            raise FileNotFoundError("Project path must be a directory")
        project_yml = path / "project.yml"
        if not project_yml.is_file():
            raise FileNotFoundError(
                "Project must contain a project.yml file at the root"
            )
        data = _load_yaml_template(project_yml)
        data["root"] = str(path)
        return cls.parse_obj(data)

    def get_profile(self, profile=None) -> dict:
        data = _load_yaml_template(self.root / self.profiles_path)
        profiles = Profiles.parse_obj(data)
        profile = profiles.default_profile if profile is None else profile
        try:
            return profiles.profiles[profile]
        except KeyError:
            raise KeyError(f"Profile {profile} is not present in profiles")

    def get_model(self) -> Model:
        data = {}
        data["tables"] = _load_dict_from_path(self.root / self.tables_path)
        data["metrics"] = _load_dict_from_path(self.root / self.metrics_path)
        data["unions"] = _load_dict_from_path(self.root / self.unions_path)
        data.update(self.dict(include={"name", "description", "locale", "currency"}))
        return Model.parse_obj(data)



from pathlib import Path

from lark import Lark, Token, Transformer, Tree

from dictum_core import utils


class Preprocessor(Transformer):
    """First pass transform. Replace // with floor and division, unquote strings etc."""

    def STRING(self, token: Token):
        """Unquotes string values (by default they are captured with '')"""
        return Token("STRING", token.value[1:-1])

    def IDENTIFIER(self, token: Token):
        return token.value

    def isnotnull(self, children):
        return Tree("NOT", [Tree("isnull", children)])

    def not_(self, children):
        return Tree("NOT", children)

    def and_(self, children):
        return Tree("AND", children)

    def or_(self, children):
        return Tree("OR", children)

    def in_(self, children):
        return Tree("IN", children)

    def not_in(self, children):
        return Tree("NOT", Tree("IN", children))

    def fdiv(self, children):
        """Replaces floor division operator (//) with
        normal division and a function call
        """
        left, right = children
        return Tree("call", ["floor", Tree("div", [left, right])])

    def call(self, children):
        """Convert all function calls to lowercase"""
        fn, *args = children
        fn = fn.lower()
        upper_fns = {"if"}  # some functions can only be uppercase methods
        fn = fn.upper() if fn in upper_fns else fn
        return Tree("call", [fn, *args])


grammar = Path(__file__).parent / "expr.lark"
parser = Lark(grammar.read_text(), start="expr", lexer="standard")
preprocessor = Preprocessor()


def parse_expr(expr: str, missing=None):
    result = preprocessor.transform(parser.parse(expr))
    if missing is not None:
        return Tree(
            "expr",
            [
                Tree(
                    "call",
                    ["coalesce", result.children[0], utils.value_to_token(missing)],
                )
            ],
        )
    return result



import pytest
from babel.localedata import locale_identifiers

from dictum_core.project.altair.locale import (
    cldr_locale_to_d3_number,
    cldr_locale_to_d3_time,
)


def test_d3_en_US_locale_number():
    assert cldr_locale_to_d3_number("en_US") == {
        "decimal": ".",
        "thousands": ",",
        "grouping": [3, 3],
        "currency": ["", ""],
    }
    assert cldr_locale_to_d3_number("en_US", "USD") == {
        "decimal": ".",
        "thousands": ",",
        "grouping": [3, 3],
        "currency": ["$", ""],
    }


def test_d3_en_RU_locale_number():
    """Just a familiar alternative locale"""
    assert cldr_locale_to_d3_number("ru_RU") == {
        "decimal": ",",
        "thousands": "\xa0",
        "grouping": [3, 3],
        "currency": ["", ""],
    }
    assert cldr_locale_to_d3_number("ru_RU", "RUB") == {
        "decimal": ",",
        "thousands": "\xa0",
        "grouping": [3, 3],
        "currency": ["", "\xa0₽"],
    }


locales = locale_identifiers()


@pytest.mark.parametrize("locale", locales)
def test_number_locales(locale: str):
    """Test that all number locales build"""
    cldr_locale_to_d3_number(locale, "USD")
    cldr_locale_to_d3_number(locale, "CNY")
    cldr_locale_to_d3_number(locale, "RUB")
    cldr_locale_to_d3_number(locale, "UAH")


@pytest.mark.parametrize("locale", locales)
def test_d3_en_US_locale_time(locale: str):
    cldr_locale_to_d3_time(locale)
